#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
# Copyright (C) 2009       B. Malengier
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Python classes
#
# -------------------------------------------------------------------------
from gi.repository import Gtk
from gi.repository import GObject
from gi.repository import GLib

# -------------------------------------------------------------------------
#
# Gramps classes
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
from gramps.gen.lib import Event, EventRef, EventRoleType, EventType
from gramps.gen.errors import WindowActiveError
from ...ddtargets import DdTargets
from .embeddedlist import TEXT_COL, MARKUP_COL, ICON_COL
from .groupembeddedlist import GroupEmbeddedList
from .eventrefmodel import EventRefModel
from ...dbguielement import DbGUIElement
from ...selectors import SelectorFactory


# -------------------------------------------------------------------------
#
# EventEmbedList
#
# -------------------------------------------------------------------------
class EventEmbedList(DbGUIElement, GroupEmbeddedList):
    _HANDLE_COL = 8
    _DND_TYPE = DdTargets.EVENTREF
    _DND_EXTRA = DdTargets.EVENT
    _WORKGROUP = EventRefModel._ROOTINDEX

    _WORKNAME = _("Family")
    _FATHNAME = _("Father")
    _MOTHNAME = _("Mother")

    _MSG = {
        "add": _("Add a new family event"),
        "del": _("Remove the selected family event"),
        "edit": _("Edit the selected family event or edit person"),
        "share": _("Share an existing event"),
        "up": _("Move the selected event upwards"),
        "down": _("Move the selected event downwards"),
    }

    # index = column in model. Value =
    #  (name, sortcol in model, width, markup/text, weigth_col
    _column_names = [
        (
            _("Description"),
            EventRefModel.COL_DESCR[0],
            240,
            TEXT_COL,
            EventRefModel.COL_FONTWEIGHT[0],
            None,
        ),
        (
            _("Type"),
            EventRefModel.COL_TYPE[0],
            120,
            TEXT_COL,
            EventRefModel.COL_FONTWEIGHT[0],
            None,
        ),
        (_("ID"), EventRefModel.COL_GID[0], 60, TEXT_COL, -1, None),
        (_("Date"), EventRefModel.COL_SORTDATE[0], 150, MARKUP_COL, -1, None),
        (_("Place"), EventRefModel.COL_PLACE[0], 240, 0, -1, None),
        (_("Role"), EventRefModel.COL_ROLE[0], 80, TEXT_COL, -1, None),
        (_("Main Participants"), EventRefModel.COL_PARTIC[0], 240, TEXT_COL, -1, None),
        None,
        None,
        None,
        (_("Age"), EventRefModel.COL_SORTAGE[0], 60, TEXT_COL, -1, None),
        None,
        (_("Private"), EventRefModel.COL_PRIVATE[0], 30, ICON_COL, -1, "gramps-lock"),
        (
            _("Source"),
            EventRefModel.COL_HAS_SOURCE[0],
            30,
            ICON_COL,
            -1,
            "gramps-source",
        ),
    ]

    def __init__(
        self,
        dbstate,
        uistate,
        track,
        obj,
        config_key,
        build_model=EventRefModel,
        **kwargs
    ):
        self.obj = obj
        self._groups = []
        self._data = []
        DbGUIElement.__init__(self, dbstate.db)
        GroupEmbeddedList.__init__(
            self,
            dbstate,
            uistate,
            track,
            _("_Events"),
            build_model,
            config_key,
            share_button=True,
            move_buttons=True,
            **kwargs,
        )

    def _connect_db_signals(self):
        """
        called on init of DbGUIElement, connect to db as required.
        """
        # note: event-rebuild closes the editors, so no need to connect to it
        self.callman.register_callbacks(
            {
                "event-update": self.event_change,  # change to an event we track
                "event-delete": self.event_delete,  # delete of event we track
            }
        )
        self.callman.connect_all(keys=["event"])

    def event_change(self, *obj):
        """
        Callback method called when a tracked event changes (description
        changes, source added, ...)
        Note that adding an event
        """
        self.rebuild_callback()

    def event_delete(self, obj):
        """
        Callback method called when a tracked event is deleted.
        There are two possibilities:
        * a tracked non-workgroup event is deleted, just rebuilding the view
            will correct this.
        * a workgroup event is deleted. The event must be removed from the obj
            so that no inconsistent data is shown.
        """
        for handle in obj:
            refs = self.get_data()[self._WORKGROUP]
            ref_list = [eref.ref for eref in refs]
            indexlist = []
            last = -1
            while True:
                try:
                    last = ref_list.index(handle, last + 1)
                    indexlist.append(last)
                except ValueError:
                    break
            # remove the deleted workgroup events from the object
            for index in reversed(indexlist):
                del refs[index]
        # now rebuild the display tab
        self.rebuild_callback()

    def get_ref_editor(self):
        from .. import EditEventRef

        return EditEventRef

    def get_icon_name(self):
        return "gramps-event"

    def get_data(self):
        # family events
        if not self._data or self.changed:
            self._data = [self.obj.get_event_ref_list()]
            self._groups = [(self.obj.get_handle(), self._WORKNAME, "")]
            # father events
            fhandle = self.obj.get_father_handle()
            if fhandle:
                fdata = self.dbstate.db.get_person_from_handle(
                    fhandle
                ).get_event_ref_list()
                if fdata:
                    self._groups.append((fhandle, self._FATHNAME, ""))
                    self._data.append(fdata)
            # mother events
            mhandle = self.obj.get_mother_handle()
            if mhandle:
                mdata = self.dbstate.db.get_person_from_handle(
                    mhandle
                ).get_event_ref_list()
                if mdata:
                    self._groups.append((mhandle, self._MOTHNAME, ""))
                    self._data.append(mdata)
            # we register all events that need to be tracked
            for group in self._data:
                self.callman.register_handles({"event": [eref.ref for eref in group]})
            self.changed = False

        return self._data

    def groups(self):
        """
        Return the (group key, group name)s in the order as given by get_data()
        """
        return self._groups

    def column_order(self):
        """
        The columns to show as a tuple containing
        tuples (show/noshow, model column)
        """
        return (
            (1, 1),  # type
            (1, 0),  # description
            (1, 3),  # date
            (1, 4),  # place
            (1, 6),  # main participants
            (1, 13),  # has source
            (1, 12),  # private
            (1, 5),  # role
            (1, 2),  # gramps id
            (1, 10),  # age
        )

    def default_types(self):
        return [
            EventType(EventType.MARRIAGE),
            EventType(EventType.DIVORCE),
        ]

    def default_type(self):
        type_list = []

        # combine return info into a single flat sequence

        event = None
        for event_ref in self.get_data()[0]:
            if event_ref.get_role() == self.default_role():
                event = self.dbstate.db.get_event_from_handle(event_ref.ref)
                type_list.append(event.get_type())

        _std_types = self.default_types()
        for etype in _std_types:
            if etype not in type_list:
                return EventType(etype)
        return _std_types[0]

    def default_role(self):
        return EventRoleType(EventRoleType.FAMILY)

    def add_button_clicked(self, obj):
        try:
            ref = EventRef()
            event = Event()
            ref.set_role(self.default_role())
            event.set_type(self.default_type())
            self.get_ref_editor()(
                self.dbstate, self.uistate, self.track, event, ref, self.object_added
            )
        except WindowActiveError:
            pass

    def __blocked_text(self):
        """
        Return the common text used when eventref cannot be edited
        """
        return _(
            "This event reference cannot be edited at this time. "
            "Either the associated event is already being edited "
            "or another event reference that is associated with "
            "the same event is being edited.\n\nTo edit this event "
            "reference, you need to close the event."
        )

    def share_button_clicked(self, obj):
        SelectEvent = SelectorFactory("Event")

        sel = SelectEvent(self.dbstate, self.uistate, self.track)
        event = sel.run()
        if event:
            try:
                ref = EventRef()
                ref.set_role(self.default_role())
                self.get_ref_editor()(
                    self.dbstate,
                    self.uistate,
                    self.track,
                    event,
                    ref,
                    self.object_added,
                )
            except WindowActiveError:
                from ...dialog import WarningDialog

                WarningDialog(
                    _("Cannot share this reference"),
                    self.__blocked_text(),
                    parent=self.uistate.window,
                )

    def edit_button_clicked(self, obj):
        ref = self.get_selected()
        if ref and ref[1] is not None and ref[0] == self._WORKGROUP:
            event = self.dbstate.db.get_event_from_handle(ref[1].ref)
            try:
                self.get_ref_editor()(
                    self.dbstate,
                    self.uistate,
                    self.track,
                    event,
                    ref[1],
                    self.object_edited,
                )
            except WindowActiveError:
                from ...dialog import WarningDialog

                WarningDialog(
                    _("Cannot edit this reference"),
                    self.__blocked_text(),
                    parent=self.uistate.window,
                )
        elif ref and ref[0] != self._WORKGROUP:
            # bring up family editor
            key = self._groups[ref[0]][0]
            self.editnotworkgroup(key)

    def object_added(self, reference, primary):
        reference.ref = primary.handle
        data = self.get_data()[self._WORKGROUP]
        data.append(reference)
        self.callman.register_handles({"event": [primary.handle]})
        self.changed = True
        self.rebuild()
        GLib.idle_add(self.tree.scroll_to_cell, (self._WORKGROUP, len(data) - 1))

    def object_edited(self, ref, event):
        """
        Called as callback after eventref has been edited.
        Note that if the event changes too (so not only the ref data), then
        an event-update signal from the database will also be raised, and the
        rebuild done here will not be needed. There is no way to avoid this ...
        """
        self.changed = True
        self.rebuild()

    def get_popup_menu_items(self):
        if self._tmpgroup == self._WORKGROUP:
            return GroupEmbeddedList.get_popup_menu_items(self)
        else:
            return [
                (True, _("_Add"), self.add_button_clicked),
                (False, _("_Edit"), self.edit_button_clicked),
            ]

    def _non_native_change(self):
        """
        handle change request of non native data
        """
        from ...dialog import WarningDialog

        WarningDialog(
            _("Cannot change Person"),
            _("You cannot change Person events in the Family Editor"),
            parent=self.uistate.window,
        )

    def _handle_drag(self, row, obj):
        """
        An event reference that is from a drag and drop has
        an unknown event reference role, so it cannot just be added,
        it needs to be edited and confirmed
        """
        if row[0] == self._WORKGROUP:
            from gramps.gen.lib import EventRoleType

            obj.set_role((EventRoleType.UNKNOWN, ""))
            # add the event
            GroupEmbeddedList._handle_drag(self, row, obj)
            # call editor to set new eventref
            event = self.dbstate.db.get_event_from_handle(obj.ref)
            try:
                self.get_ref_editor()(
                    self.dbstate,
                    self.uistate,
                    self.track,
                    event,
                    obj,
                    self.object_edited,
                )
            except WindowActiveError:
                from ...dialog import WarningDialog

                WarningDialog(
                    _("Cannot edit this reference"),
                    _(
                        "This event reference cannot be edited at this time. "
                        "Either the associated event is already being edited "
                        "or another event reference that is associated with "
                        "the same event is being edited.\n\nTo edit this event "
                        "reference, you need to close the event."
                    ),
                    parent=self.uistate.window,
                )
        else:
            self.dropnotworkgroup(row, obj)

    def handle_extra_type(self, objtype, obj):
        try:
            ref = EventRef()
            event = self.dbstate.db.get_event_from_handle(obj)
            ref.set_role(self.default_role())
            self.get_ref_editor()(
                self.dbstate, self.uistate, self.track, event, ref, self.object_added
            )
        except WindowActiveError:
            pass

    def editnotworkgroup(self, key):
        """
        Edit non native event in own editor
        """
        person = self.dbstate.db.get_person_from_handle(key)
        try:
            from .. import EditPerson

            EditPerson(self.dbstate, self.uistate, [], person)
        except WindowActiveError:
            pass

    def dropnotworkgroup(self, row, obj):
        """
        Drop of obj on row that is not WORKGROUP
        """
        self._non_native_change()

    def move_away_work(self, row_from, row_to, obj):
        """
        move from the workgroup to a not workgroup
        we allow to change the default name like this
        """
        self.dropnotworkgroup(None, None)

    def move_to_work(self, row_from, row_to, obj):
        """
        move from a non workgroup to the workgroup
        handle this as a drop from clipboard
        """
        self._handle_drag(row_to, obj)

    def _move_up_notwork(self, row_from, obj, selmethod=None):
        """
        move up outside of workgroup
        """
        self._non_native_change()

    def _move_down_notwork(self, row_from, obj, selmethod=None):
        """
        move up outside of workgroup
        """
        self._non_native_change()

    def post_rebuild(self, prebuildpath):
        """
        Allow post rebuild specific handling.
        @param prebuildpath: path selected before rebuild, None if none
        @type prebuildpath: tree path
        """
        self.tree.expand_all()
        if prebuildpath is not None:
            self.selection.select_path(prebuildpath)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2010,2017 Nick Hall
# Copyright (C) 2013      Doug Blank <doug.blank@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Table Object class for Gramps.
"""

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import time
from abc import abstractmethod

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..const import GRAMPS_LOCALE as glocale
from .baseobj import BaseObject

CODESET = glocale.encoding


# -------------------------------------------------------------------------
#
# TableObject
#
# -------------------------------------------------------------------------
class TableObject(BaseObject):
    """
    The TableObject is the base class for all objects that are stored in a
    seperate database table.  Each object has a database handle and a last
    changed time.  The database handle is used as the unique key for a record
    in the database.  This is not the same as the Gramps ID, which is a user
    visible identifier for a record.

    It is the base class for the BasicPrimaryObject class and Tag class.
    """

    def __init__(self, source=None):
        """
        Initialize a TableObject.

        If source is None, the handle is assigned as an empty string.
        If source is not None, then the handle is initialized from the value in
        the source object.

        :param source: Object used to initialize the new object
        :type source: TableObject
        """
        if source:
            self.handle = source.handle
            self.change = source.change
        else:
            self.handle = None
            self.change = 0

    @abstractmethod
    def serialize(self):
        """
        Convert the object to a serialized tuple of data.
        """

    @abstractmethod
    def unserialize(self, data):
        """
        Convert a serialized tuple of data to an object.
        """

    def get_change_time(self):
        """
        Return the time that the data was last changed.

        The value in the format returned by the :meth:`time.time()` command.

        :returns: Time that the data was last changed. The value in the format
                  returned by the :meth:`time.time()` command.
        :rtype: int
        """
        return self.change

    def set_change_time(self, change):
        """
        Modify the time that the data was last changed.

        The value must be in the format returned by the :meth:`time.time()`
        command.

        :param change: new time
        :type change: int in format as :meth:`time.time()` command
        """
        self.change = change

    def get_change_display(self):
        """
        Return the string representation of the last change time.

        :returns: string representation of the last change time.
        :rtype: str

        """
        if self.change:
            return str(time.strftime("%x %X", time.localtime(self.change)), CODESET)
        return ""

    def set_handle(self, handle):
        """
        Set the database handle for the primary object.

        :param handle: object database handle
        :type handle: str
        """
        self.handle = handle

    def get_handle(self):
        """
        Return the database handle for the primary object.

        :returns: database handle associated with the object
        :rtype: str
        """
        return self.handle

    @classmethod
    def get_schema(cls):
        """
        Return schema.
        """
        return {}

    @classmethod
    def get_secondary_fields(cls):
        """
        Return all secondary fields and their types
        """
        result = []
        for key, value in cls.get_schema()["properties"].items():
            schema_type = value.get("type")
            if isinstance(schema_type, list):
                schema_type.remove("null")
                schema_type = schema_type[0]
            elif isinstance(schema_type, dict):
                schema_type = None
            if schema_type in ("string", "integer", "number", "boolean"):
                result.append((key.lower(), schema_type, value.get("maxLength")))
        return result



# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2001-2006  Donald N. Allingham
# Copyright (C) 2008       Gary Burton
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Citation List View
"""

# -------------------------------------------------------------------------
#
# python modules
#
# -------------------------------------------------------------------------
import logging

LOG = logging.getLogger(".citation")

# -------------------------------------------------------------------------
#
# GTK/Gnome modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# gramps modules
#
# -------------------------------------------------------------------------
from gramps.gui.views.treemodels.citationlistmodel import CitationListModel
from gramps.gen.plug import CATEGORY_QR_CITATION
from gramps.gen.lib import Citation, Source
from gramps.gui.views.listview import ListView, TEXT, MARKUP, ICON
from gramps.gui.views.bookmarks import CitationBookmarks
from gramps.gen.errors import WindowActiveError
from gramps.gui.ddtargets import DdTargets
from gramps.gui.dialog import ErrorDialog
from gramps.gui.editors import EditCitation
from gramps.gui.filters.sidebar import CitationSidebarFilter
from gramps.gui.merge import MergeCitation

# -------------------------------------------------------------------------
#
# internationalization
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext


# -------------------------------------------------------------------------
#
# CitationView
#
# -------------------------------------------------------------------------
class CitationListView(ListView):
    """
    A list view of citations.

    The citation list view only shows the citations (it does not show
    sources as separate list entries).
    """

    # The data items here have to correspond, in order, to the items in
    # src/giu/views/treemodels/citationlismodel.py
    COL_TITLE_PAGE = 0
    COL_ID = 1
    COL_DATE = 2
    COL_CONFIDENCE = 3
    COL_PRIV = 4
    COL_TAGS = 5
    COL_CHAN = 6
    COL_SRC_TITLE = 7
    COL_SRC_ID = 8
    COL_SRC_AUTH = 9
    COL_SRC_ABBR = 10
    COL_SRC_PINFO = 11
    COL_SRC_PRIV = 12
    COL_SRC_CHAN = 13
    # column definitions
    COLUMNS = [
        (_("Volume/Page"), TEXT, None),
        (_("ID"), TEXT, None),
        (_("Date"), MARKUP, None),
        (_("Confidence"), TEXT, None),
        (_("Private"), ICON, "gramps-lock"),
        (_("Tags"), TEXT, None),
        (_("Last Changed"), TEXT, None),
        (_("Source: Title"), TEXT, None),
        (_("Source: ID"), TEXT, None),
        (_("Source: Author"), TEXT, None),
        (_("Source: Abbreviation"), TEXT, None),
        (_("Source: Publication Information"), TEXT, None),
        (_("Source: Private"), ICON, "gramps-lock"),
        (_("Source: Last Changed"), TEXT, None),
    ]
    # default setting with visible columns, order of the col, and their size
    CONFIGSETTINGS = (
        ("columns.visible", [COL_TITLE_PAGE, COL_ID, COL_DATE, COL_CONFIDENCE]),
        (
            "columns.rank",
            [
                COL_TITLE_PAGE,
                COL_ID,
                COL_DATE,
                COL_CONFIDENCE,
                COL_PRIV,
                COL_TAGS,
                COL_CHAN,
                COL_SRC_TITLE,
                COL_SRC_ID,
                COL_SRC_AUTH,
                COL_SRC_ABBR,
                COL_SRC_PINFO,
                COL_SRC_PRIV,
                COL_SRC_CHAN,
            ],
        ),
        (
            "columns.size",
            [200, 75, 100, 100, 40, 100, 100, 200, 75, 75, 100, 150, 40, 100],
        ),
    )
    ADD_MSG = _("Add a new citation and a new source")
    ADD_SOURCE_MSG = _("Add a new source")
    ADD_CITATION_MSG = _("Add a new citation to an existing source")
    EDIT_MSG = _("Edit the selected citation")
    DEL_MSG = _("Delete the selected citation")
    MERGE_MSG = _("Merge the selected citations")
    FILTER_TYPE = "Citation"
    QR_CATEGORY = CATEGORY_QR_CITATION

    def __init__(self, pdata, dbstate, uistate, nav_group=0):
        signal_map = {
            "citation-add": self.row_add,
            "citation-update": self.row_update,
            "citation-delete": self.row_delete,
            "citation-rebuild": self.object_build,
        }

        ListView.__init__(
            self,
            _("Citation View"),
            pdata,
            dbstate,
            uistate,
            CitationListModel,
            signal_map,
            CitationBookmarks,
            nav_group,
            multiple=True,
            filter_class=CitationSidebarFilter,
        )

        self.additional_uis.append(self.additional_ui)

    def navigation_type(self):
        return "Citation"

    def drag_info(self):
        return DdTargets.CITATION_LINK

    def get_stock(self):
        return "gramps-citation"

    additional_ui = [
        """
      <placeholder id="LocalExport">
        <item>
          <attribute name="action">win.ExportTab</attribute>
          <attribute name="label" translatable="yes">Export View...</attribute>
        </item>
      </placeholder>
""",
        """
      <section id="AddEditBook">
        <item>
          <attribute name="action">win.AddBook</attribute>
          <attribute name="label" translatable="yes">_Add Bookmark</attribute>
        </item>
        <item>
          <attribute name="action">win.EditBook</attribute>
          <attribute name="label" translatable="no">%s...</attribute>
        </item>
      </section>
"""
        % _("Organize Bookmarks"),
        """
      <placeholder id="CommonGo">
      <section>
        <item>
          <attribute name="action">win.Back</attribute>
          <attribute name="label" translatable="yes">_Back</attribute>
        </item>
        <item>
          <attribute name="action">win.Forward</attribute>
          <attribute name="label" translatable="yes">_Forward</attribute>
        </item>
      </section>
      </placeholder>
""",
        """
      <section id='CommonEdit' groups='RW'>
        <item>
          <attribute name="action">win.Add</attribute>
          <attribute name="label" translatable="yes">_Add...</attribute>
        </item>
        <item>
          <attribute name="action">win.Edit</attribute>
          <attribute name="label">%s</attribute>
        </item>
        <item>
          <attribute name="action">win.Remove</attribute>
          <attribute name="label" translatable="yes">_Delete</attribute>
        </item>
        <item>
          <attribute name="action">win.Merge</attribute>
          <attribute name="label" translatable="yes">_Merge...</attribute>
        </item>
      </section>
"""
        % _("_Edit...", "action"),  # to use sgettext()
        """
        <placeholder id='otheredit'>
        <item>
          <attribute name="action">win.FilterEdit</attribute>
          <attribute name="label" translatable="yes">"""
        """Citation Filter Editor</attribute>
        </item>
        </placeholder>
""",  # Following are the Toolbar items
        """
    <placeholder id='CommonNavigation'>
    <child groups='RO'>
      <object class="GtkToolButton">
        <property name="icon-name">go-previous</property>
        <property name="action-name">win.Back</property>
        <property name="tooltip_text" translatable="yes">"""
        """Go to the previous object in the history</property>
        <property name="label" translatable="yes">_Back</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    <child groups='RO'>
      <object class="GtkToolButton">
        <property name="icon-name">go-next</property>
        <property name="action-name">win.Forward</property>
        <property name="tooltip_text" translatable="yes">"""
        """Go to the next object in the history</property>
        <property name="label" translatable="yes">_Forward</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    </placeholder>
""",
        """
    <placeholder id='BarCommonEdit'>
    <child groups='RW'>
      <object class="GtkToolButton">
        <property name="icon-name">list-add</property>
        <property name="action-name">win.Add</property>
        <property name="tooltip_text">%s</property>
        <property name="label" translatable="yes">_Add...</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    <child groups='RW'>
      <object class="GtkToolButton">
        <property name="icon-name">gtk-edit</property>
        <property name="action-name">win.Edit</property>
        <property name="tooltip_text">%s</property>
        <property name="label" translatable="yes">Edit...</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    <child groups='RW'>
      <object class="GtkToolButton">
        <property name="icon-name">list-remove</property>
        <property name="action-name">win.Remove</property>
        <property name="tooltip_text">%s</property>
        <property name="label" translatable="yes">_Delete</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    <child groups='RW'>
      <object class="GtkToolButton">
        <property name="icon-name">gramps-merge</property>
        <property name="action-name">win.Merge</property>
        <property name="tooltip_text" >%s</property>
        <property name="label" translatable="yes">_Merge...</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    </placeholder>
"""
        % (ADD_MSG, EDIT_MSG, DEL_MSG, MERGE_MSG),
        """
    <menu id="Popup">
      <section>
        <item>
          <attribute name="action">win.Back</attribute>
          <attribute name="label" translatable="yes">_Back</attribute>
        </item>
        <item>
          <attribute name="action">win.Forward</attribute>
          <attribute name="label" translatable="yes">Forward</attribute>
        </item>
      </section>
      <section id="PopUpTree">
      </section>
      <section>
        <item>
          <attribute name="action">win.Add</attribute>
          <attribute name="label" translatable="yes">_Add...</attribute>
        </item>
        <item>
          <attribute name="action">win.Edit</attribute>
          <attribute name="label">%s</attribute>
        </item>
        <item>
          <attribute name="action">win.Remove</attribute>
          <attribute name="label" translatable="yes">_Delete</attribute>
        </item>
        <item>
          <attribute name="action">win.Merge</attribute>
          <attribute name="label" translatable="yes">_Merge...</attribute>
        </item>
      </section>
      <section>
        <placeholder id='QuickReport'>
        </placeholder>
        <placeholder id='WebConnect'>
        </placeholder>
      </section>
    </menu>
"""
        % _("_Edit...", "action"),  # to use sgettext()
    ]

    def add(self, *obj):
        """
        add:        Add a new citation and a new source (this can also be done
                      by source view add a source, then citation view add a new
                      citation to an existing source)

        Create a new Source instance and Citation instance and call the
        EditCitation editor with the new source and new citation.

        Called when the Add button is clicked.
        If the window already exists (WindowActiveError), we ignore it.
        This prevents the dialog from coming up twice on the same object.

        However, since the window is identified by the Source object, and
        we have just created a new one, it seems to be impossible for the
        window to already exist, so this is just an extra safety measure.
        """
        try:
            EditCitation(self.dbstate, self.uistate, [], Citation(), Source())
        except WindowActiveError:
            pass

    def remove(self, *obj):
        handles = self.selected_handles()
        ht_list = [("Citation", hndl) for hndl in handles]
        self.remove_selected_objects(ht_list)

    def edit(self, *obj):
        """
        Edit a Citation
        """
        for handle in self.selected_handles():
            citation = self.dbstate.db.get_citation_from_handle(handle)
            try:
                EditCitation(self.dbstate, self.uistate, [], citation)
            except WindowActiveError:
                pass

    def __blocked_text(self):
        """
        Return the common text used when citation cannot be edited
        """
        return _(
            "This citation cannot be edited at this time. "
            "Either the associated citation is already being "
            "edited or another object that is associated with "
            "the same citation is being edited.\n\nTo edit this "
            "citation, you need to close the object."
        )

    def merge(self, *obj):
        """
        Merge the selected citations.
        """
        mlist = self.selected_handles()

        if len(mlist) != 2:
            msg = _("Cannot merge citations.")
            msg2 = _(
                "Exactly two citations must be selected to perform a "
                "merge. A second citation can be selected by holding "
                "down the control key while clicking on the desired "
                "citation."
            )
            ErrorDialog(msg, msg2, parent=self.uistate.window)
        else:
            citation1 = self.dbstate.db.get_citation_from_handle(mlist[0])
            citation2 = self.dbstate.db.get_citation_from_handle(mlist[1])
            if not citation1.get_reference_handle() == citation2.get_reference_handle():
                msg = _("Cannot merge citations.")
                msg2 = _(
                    "The two selected citations must have the same "
                    "source to perform a merge. If you want to merge "
                    "these two citations, then you must merge the "
                    "sources first."
                )
                ErrorDialog(msg, msg2, parent=self.uistate.window)
            else:
                MergeCitation(self.dbstate, self.uistate, [], mlist[0], mlist[1])

    def get_handle_from_gramps_id(self, gid):
        obj = self.dbstate.db.get_citation_from_gramps_id(gid)
        if obj:
            return obj.get_handle()
        else:
            return None

    def tag_updated(self, handle_list):
        """
        Update tagged rows when a tag color changes.
        """
        all_links = set([])
        for tag_handle in handle_list:
            links = set(
                [
                    link[1]
                    for link in self.dbstate.db.find_backlink_handles(
                        tag_handle, include_classes="Citation"
                    )
                ]
            )
            all_links = all_links.union(links)
        self.row_update(list(all_links))

    def add_tag(self, transaction, citation_handle, tag_handle):
        """
        Add the given tag to the given citation.
        """
        citation = self.dbstate.db.get_citation_from_handle(citation_handle)
        citation.add_tag(tag_handle)
        self.dbstate.db.commit_citation(citation, transaction)

    def remove_tag(self, transaction, citation_handle, tag_handle):
        """
        Remove the given tag from the given citation.
        """
        citation = self.dbstate.db.get_citation_from_handle(citation_handle)
        citation.remove_tag(tag_handle)
        self.dbstate.db.commit_citation(citation, transaction)

    def get_default_gramplets(self):
        """
        Define the default gramplets for the sidebar and bottombar.
        This is overridden for the tree view to give 'Source Filter'
        """
        return (
            ("Citation Filter",),
            ("Citation Gallery", "Citation Notes", "Citation Backlinks"),
        )

    def get_config_name(self):
        return __name__



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2004-2007  Donald N. Allingham
# Copyright (C) 2010       Brian G. Matherly
# Copyright (C) 2014       Paul Franklin
# Copyright (C) 2024       Doug Blank
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Class handling language-specific displaying of names.

Specific symbols for parts of a name are defined:

    ======  ===============================================================
    Symbol  Description
    ======  ===============================================================
    't'     title
    'f'     given (first names)
    'l'     full surname (lastname)
    'c'     callname
    'x'     nick name, call, or otherwise first first name (common name)
    'i'     initials of the first names
    'm'     primary surname (main)
    '0m'    primary surname prefix
    '1m'    primary surname surname
    '2m'    primary surname connector
    'y'     pa/matronymic surname (father/mother) - assumed unique
    '0y'    pa/matronymic prefix
    '1y'    pa/matronymic surname
    '2y'    pa/matronymic connector
    'o'     surnames without pa/matronymic and primary
    'r'     non primary surnames (rest)
    'p'     list of all prefixes
    'q'     surnames without prefixes and connectors
    's'     suffix
    'n'     nick name
    'g'     family nick name
    ======  ===============================================================
"""

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
from __future__ import annotations
import re
import logging
from typing import Callable

LOG = logging.getLogger(".gramps.gen")

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..const import ARABIC_COMMA, ARABIC_SEMICOLON, GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext
from ..lib.name import Name
from ..lib.nameorigintype import NameOriginType
from ..lib.json_utils import object_to_data

try:
    from ..config import config

    WITH_GRAMPS_CONFIG = True
except ImportError:
    WITH_GRAMPS_CONFIG = False


# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------
_ORIGINPATRO = NameOriginType.PATRONYMIC
_ORIGINMATRO = NameOriginType.MATRONYMIC
_ACT = True
_INA = False

_F_NAME = 0  # name of the format
_F_FMT = 1  # the format string
_F_ACT = 2  # if the format is active
_F_FN = 3  # name format function
_F_RAWFN = 4  # name format raw function

PAT_AS_SURN = False


# -------------------------------------------------------------------------
#
# Local functions
#
# -------------------------------------------------------------------------
# Because of occurring in an exec(), this couldn't be in a lambda:
# we sort names first on longest first, then last letter first, this to
# avoid translations of shorter terms which appear in longer ones, eg
# namelast may not be mistaken with name, so namelast must first be
# converted to %k before name is converted.
##def _make_cmp(a, b): return -cmp((len(a[1]),a[1]), (len(b[1]), b[1]))
def _make_cmp_key(a):
    return (len(a[1]), a[1])  # set reverse to True!!


# -------------------------------------------------------------------------
#
# NameDisplayError class
#
# -------------------------------------------------------------------------
class NameDisplayError(Exception):
    """
    Error used to report that the name display format string is invalid.
    """

    def __init__(self, value):
        Exception.__init__(self)
        self.value = value

    def __str__(self):
        return self.value


# -------------------------------------------------------------------------
#
# Functions to extract data from raw lists (unserialized objects)
#
# -------------------------------------------------------------------------


def _raw_full_surname(raw_surn_data_list):
    """method for the 'l' symbol: full surnames"""
    result = ""
    for raw_surn_data in raw_surn_data_list:
        result += __format_raw_surname(raw_surn_data)
    return result.strip()


def _raw_primary_surname(raw_surn_data_list):
    """method for the 'm' symbol: primary surname"""
    global PAT_AS_SURN
    nrsur = len(raw_surn_data_list)
    for raw_surn_data in raw_surn_data_list:
        if raw_surn_data["primary"]:
            # if there are multiple surnames, return the primary. If there
            # is only one surname, then primary has little meaning, and we
            # assume a pa/matronymic should not be given as primary as it
            # normally is defined independently
            if (
                not PAT_AS_SURN
                and nrsur == 1
                and (
                    raw_surn_data["origintype"]["value"] == _ORIGINPATRO
                    or raw_surn_data["origintype"]["value"] == _ORIGINMATRO
                )
            ):
                return ""
            else:
                return __format_raw_surname(raw_surn_data).strip()
    return ""


def _raw_primary_surname_only(raw_surn_data_list):
    """method to obtain the raw primary surname data, so this returns a string"""
    global PAT_AS_SURN
    nrsur = len(raw_surn_data_list)
    for raw_surn_data in raw_surn_data_list:
        if raw_surn_data["primary"]:
            if (
                not PAT_AS_SURN
                and nrsur == 1
                and (
                    raw_surn_data["origintype"]["value"] == _ORIGINPATRO
                    or raw_surn_data["origintype"]["value"] == _ORIGINMATRO
                )
            ):
                return ""
            else:
                return raw_surn_data["surname"]
    return ""


def _raw_primary_prefix_only(raw_surn_data_list):
    """method to obtain the raw primary surname data"""
    global PAT_AS_SURN
    nrsur = len(raw_surn_data_list)
    for raw_surn_data in raw_surn_data_list:
        if raw_surn_data["primary"]:
            if (
                not PAT_AS_SURN
                and nrsur == 1
                and (
                    raw_surn_data["origintype"]["value"] == _ORIGINPATRO
                    or raw_surn_data["origintype"]["value"] == _ORIGINMATRO
                )
            ):
                return ""
            else:
                return raw_surn_data["prefix"]
    return ""


def _raw_primary_conn_only(raw_surn_data_list):
    """method to obtain the raw primary surname data"""
    global PAT_AS_SURN
    nrsur = len(raw_surn_data_list)
    for raw_surn_data in raw_surn_data_list:
        if raw_surn_data["primary"]:
            if (
                not PAT_AS_SURN
                and nrsur == 1
                and (
                    raw_surn_data["origintype"]["value"] == _ORIGINPATRO
                    or raw_surn_data["origintype"]["value"] == _ORIGINMATRO
                )
            ):
                return ""
            else:
                return raw_surn_data["connector"]
    return ""


def _raw_patro_surname(raw_surn_data_list):
    """method for the 'y' symbol: patronymic surname"""
    for raw_surn_data in raw_surn_data_list:
        if (
            raw_surn_data["origintype"]["value"] == _ORIGINPATRO
            or raw_surn_data["origintype"]["value"] == _ORIGINMATRO
        ):
            return __format_raw_surname(raw_surn_data).strip()
    return ""


def _raw_patro_surname_only(raw_surn_data_list):
    """method for the '1y' symbol: patronymic surname only"""
    for raw_surn_data in raw_surn_data_list:
        if (
            raw_surn_data["origintype"]["value"] == _ORIGINPATRO
            or raw_surn_data["origintype"]["value"] == _ORIGINMATRO
        ):
            result = "%s" % (raw_surn_data["surname"])
            return " ".join(result.split())
    return ""


def _raw_patro_prefix_only(raw_surn_data_list):
    """method for the '0y' symbol: patronymic prefix only"""
    for raw_surn_data in raw_surn_data_list:
        if (
            raw_surn_data["origintype"]["value"] == _ORIGINPATRO
            or raw_surn_data["origintype"]["value"] == _ORIGINMATRO
        ):
            result = "%s" % (raw_surn_data["prefix"])
            return " ".join(result.split())
    return ""


def _raw_patro_conn_only(raw_surn_data_list):
    """method for the '2y' symbol: patronymic conn only"""
    for raw_surn_data in raw_surn_data_list:
        if (
            raw_surn_data["origintype"]["value"] == _ORIGINPATRO
            or raw_surn_data["origintype"]["value"] == _ORIGINMATRO
        ):
            result = "%s" % (raw_surn_data["connector"])
            return " ".join(result.split())
    return ""


def _raw_nonpatro_surname(raw_surn_data_list):
    """method for the 'o' symbol: full surnames without pa/matronymic or
    primary
    """
    result = ""
    for raw_surn_data in raw_surn_data_list:
        if (
            (not raw_surn_data["primary"])
            and raw_surn_data["origintype"]["value"] != _ORIGINPATRO
            and raw_surn_data["origintype"]["value"] != _ORIGINMATRO
        ):
            result += __format_raw_surname(raw_surn_data)
    return result.strip()


def _raw_nonprimary_surname(raw_surn_data_list):
    """method for the 'r' symbol: nonprimary surnames"""
    result = ""
    for raw_surn_data in raw_surn_data_list:
        if not raw_surn_data["primary"]:
            result += __format_raw_surname(raw_surn_data)
    return result.strip()


def _raw_prefix_surname(raw_surn_data_list):
    """method for the 'p' symbol: all prefixes"""
    result = ""
    for raw_surn_data in raw_surn_data_list:
        result += "%s " % (raw_surn_data["prefix"])
    return " ".join(result.split()).strip()


def _raw_single_surname(raw_surn_data_list):
    """method for the 'q' symbol: surnames without prefix and connectors"""
    result = ""
    for raw_surn_data in raw_surn_data_list:
        result += "%s " % (raw_surn_data["surname"])
    return " ".join(result.split()).strip()


def cleanup_name(namestring):
    """Remove too long white space due to missing name parts,
    so "a   b" becomes "a b" and "a , b" becomes "a, b"
    """
    parts = namestring.split()
    if not parts:
        return ""
    result = parts[0]
    for val in parts[1:]:
        if len(val) == 1 and val in [",", ";", ":", ARABIC_COMMA, ARABIC_SEMICOLON]:
            result += val
        else:
            result += " " + val

    return result


def __format_raw_surname(raw_surn_data):
    """
    Return a formatted string representing one surname part.

    If the connector is a hyphen, don't pad it with spaces.
    """
    result = raw_surn_data["prefix"]
    if result:
        result += " "
    result += raw_surn_data["surname"]
    if result and raw_surn_data["connector"] != "-":
        result += " %s " % raw_surn_data["connector"]
    else:
        result += raw_surn_data["connector"]
    return result


# -------------------------------------------------------------------------
#
# NameDisplay class
#
# -------------------------------------------------------------------------
class NameDisplay:
    """
    Base class for displaying of Name instances.

    Property:
      *default_format*
        the default name format to use
      *pas_as_surn*
        if only one surname, see if pa/ma should be considered as 'the' surname.
    """

    format_funcs: dict[str, Callable] = {}
    raw_format_funcs: dict[str, Callable] = {}

    def __init__(self, xlocale=glocale):
        """
        Initialize the NameDisplay class.

        If xlocale is passed in (a GrampsLocale), then
        the translated script will be returned instead.

        :param xlocale: allow selection of the displayer script
        :type xlocale: a GrampsLocale instance
        """
        global WITH_GRAMPS_CONFIG
        global PAT_AS_SURN

        # Translators: needed for Arabic, ignore otherwise
        COMMAGLYPH = xlocale.translation.gettext(",")

        self.STANDARD_FORMATS = [
            (Name.DEF, _("Default format (defined by Gramps preferences)"), "", _ACT),
            (Name.LNFN, _("Surname, Given Suffix"), "%l" + COMMAGLYPH + " %f %s", _ACT),
            (Name.FN, _("Given"), "%f", _ACT),
            (Name.FNLN, _("Given Surname Suffix"), "%f %l %s", _ACT),
            # primary name primconnector other, given pa/matronynic suffix, primprefix
            # Translators: long string, have a look at Preferences dialog
            (
                Name.LNFNP,
                _("Main Surnames, Given Patronymic Suffix Prefix"),
                "%1m %2m %o" + COMMAGLYPH + " %f %1y %s %0m",
                _ACT,
            ),
            # DEPRECATED FORMATS
            (Name.PTFN, _("Patronymic, Given"), "%y" + COMMAGLYPH + " %s %f", _INA),
        ]

        self.LNFN_STR = "%s" + COMMAGLYPH + " %s %s"

        self.name_formats = {}

        if WITH_GRAMPS_CONFIG:
            self.default_format = config.get("preferences.name-format")
            if self.default_format == 0:
                self.default_format = Name.LNFN
                config.set("preferences.name-format", self.default_format)
            # if only one surname, see if pa/ma should be considered as
            # 'the' surname.
            PAT_AS_SURN = config.get("preferences.patronimic-surname")
            config.connect("preferences.patronimic-surname", self.change_pa_sur)
        else:
            self.default_format = Name.LNFN
            PAT_AS_SURN = False

        # preinit the name formats, this should be updated with the data
        # in the database once a database is loaded
        self.set_name_format(self.STANDARD_FORMATS)

    def change_pa_sur(self, *args):
        """How to handle single patronymic as surname is changed"""
        global PAT_AS_SURN
        PAT_AS_SURN = config.get("preferences.patronimic-surname")

    def get_pat_as_surn(self):
        global PAT_AS_SURN
        return PAT_AS_SURN

    def _format_fn(self, fmt_str):
        return lambda x: self.format_str(x, fmt_str)

    def _format_raw_fn(self, fmt_str):
        return lambda x: self.format_str_raw(x, fmt_str)

    def _raw_lnfn(self, raw_data):
        result = self.LNFN_STR % (
            _raw_full_surname(raw_data["surname_list"]),
            raw_data["first_name"],
            raw_data["suffix"],
        )
        return " ".join(result.split())

    def _raw_fnln(self, raw_data):
        result = "%s %s %s" % (
            raw_data["first_name"],
            _raw_full_surname(raw_data["surname_list"]),
            raw_data["suffix"],
        )
        return " ".join(result.split())

    def _raw_fn(self, raw_data):
        result = raw_data["first_name"]
        return " ".join(result.split())

    def clear_custom_formats(self):
        self.name_formats = {
            num: value for num, value in self.name_formats.items() if num >= 0
        }

    def set_name_format(self, formats):
        raw_func_dict = {
            Name.LNFN: self._raw_lnfn,
            Name.FNLN: self._raw_fnln,
            Name.FN: self._raw_fn,
        }

        for num, name, fmt_str, act in formats:
            func = self._format_fn(fmt_str)
            func_raw = raw_func_dict.get(num, self._format_raw_fn(fmt_str))
            self.name_formats[num] = (name, fmt_str, act, func, func_raw)
        self.set_default_format(self.get_default_format())

    def add_name_format(self, name, fmt_str):
        for num in self.name_formats:
            if fmt_str in self.name_formats.get(num):
                return num
        num = -1
        while num in self.name_formats:
            num -= 1
        self.set_name_format([(num, name, fmt_str, _ACT)])
        return num

    def edit_name_format(self, num, name, fmt_str):
        self.set_name_format([(num, name, fmt_str, _ACT)])
        if self.default_format == num:
            self.set_default_format(num)

    def del_name_format(self, num):
        try:
            del self.name_formats[num]
        except:
            pass

    def set_default_format(self, num):
        if num not in self.name_formats:
            num = Name.LNFN
        # if user sets default format to the Gramps default format,
        # then we select LNFN as format.
        if num == Name.DEF:
            num = Name.LNFN

        self.default_format = num

        self.name_formats[Name.DEF] = (
            self.name_formats[Name.DEF][_F_NAME],
            self.name_formats[Name.DEF][_F_FMT],
            self.name_formats[Name.DEF][_F_ACT],
            self.name_formats[num][_F_FN],
            self.name_formats[num][_F_RAWFN],
        )

    def get_default_format(self):
        return self.default_format

    def set_format_inactive(self, num):
        try:
            self.name_formats[num] = (
                self.name_formats[num][_F_NAME],
                self.name_formats[num][_F_FMT],
                _INA,
                self.name_formats[num][_F_FN],
                self.name_formats[num][_F_RAWFN],
            )
        except:
            pass

    def get_name_format(self, also_default=False, only_custom=False, only_active=True):
        """
        Returns a list of name formats as tuples on
        the form (index, name,fmt_str,act).
        The will contain standard formats followed
        by custom formats both in ascending order on
        their indices.
        """

        custom_formats = sorted(
            [
                (index, name, format_string, active)
                for index, (
                    name,
                    format_string,
                    active,
                    *rest,
                ) in self.name_formats.items()
                if index < 0 and (not only_active or active)
            ]
        )

        if only_custom:
            return custom_formats

        standard_formats = sorted(
            [
                (index, name, format_string, active)
                for index, (
                    name,
                    format_string,
                    active,
                    *rest,
                ) in self.name_formats.items()
                if index >= 0
                and (also_default or index)
                and (not only_active or active)
            ]
        )

        return standard_formats + custom_formats

    def _is_format_valid(self, num):
        try:
            if not self.name_formats[num][_F_ACT]:
                num = 0
        except:
            num = 0
        return num

    # -------------------------------------------------------------------------

    def _gen_raw_func(self, format_str):
        """The job of building the name from a format string is rather
        expensive and it is called lots and lots of times. So it is worth
        going to some length to optimise it as much as possible.

        This method constructs a new function that is specifically written
        to format a name given a particular format string. This is worthwhile
        because the format string itself rarely changes, so by caching the new
        function and calling it directly when asked to format a name to the
        same format string again we can be as quick as possible.

        The new function is of the form::

        def fn(raw_data):
            return "%s %s %s" % (raw_data["title"],
                   raw_data["first_name"],
                   raw_data["suffix"])

        Specific symbols for parts of a name are defined (keywords given):
        't' : title = title
        'f' : given = given (first names)
        'l' : surname = full surname (lastname)
        'c' : call = callname
        'x' : common = nick name, call, otherwise first first name (common name)
        'i' : initials = initials of the first names
        'm' : primary = primary surname (main)
        '0m': primary[pre]= prefix primary surname (main)
        '1m': primary[sur]= surname primary surname (main)
        '2m': primary[con]= connector primary surname (main)
        'y' : patronymic = pa/matronymic surname (father/mother) - assumed unique
        '0y': patronymic[pre] = prefix      "
        '1y': patronymic[sur] = surname     "
        '2y': patronymic[con] = connector   "
        'o' : notpatronymic = surnames without pa/matronymic and primary
        'r' : rest = non primary surnames
        'p' : prefix = list of all prefixes
        'q' : rawsurnames = surnames without prefixes and connectors
        's' : suffix = suffix
        'n' : nickname = nick name
        'g' : familynick = family nick name


        """

        # we need the names of each of the variables or methods that are
        # called to fill in each format flag.
        # Dictionary is "code": ("expression", "keyword", "i18n-keyword")
        d = {
            "t": ("raw_data['title']", "title", _("title", "Person")),
            "f": ("raw_data['first_name']", "given", _("given")),
            "l": (
                "_raw_full_surname(raw_data['surname_list'])",
                "surname",
                _("surname"),
            ),
            "s": ("raw_data['suffix']", "suffix", _("suffix")),
            "c": ("raw_data['call']", "call", _("call", "Name")),
            "x": (
                "(raw_data['nick'] or raw_data['call'] or raw_data['first_name'].split(' ')[0])",
                "common",
                _("common", "Name"),
            ),
            "i": (
                "''.join([word[0] +'.' for word in ('. ' +"
                + " raw_data['first_name']).split()][1:])",
                "initials",
                _("initials"),
            ),
            "m": (
                "_raw_primary_surname(raw_data['surname_list'])",
                "primary",
                _("primary", "Name"),
            ),
            "0m": (
                "_raw_primary_prefix_only(raw_data['surname_list'])",
                "primary[pre]",
                _("primary[pre]"),
            ),
            "1m": (
                "_raw_primary_surname_only(raw_data['surname_list'])",
                "primary[sur]",
                _("primary[sur]"),
            ),
            "2m": (
                "_raw_primary_conn_only(raw_data['surname_list'])",
                "primary[con]",
                _("primary[con]"),
            ),
            "y": (
                "_raw_patro_surname(raw_data['surname_list'])",
                "patronymic",
                _("patronymic"),
            ),
            "0y": (
                "_raw_patro_prefix_only(raw_data['surname_list'])",
                "patronymic[pre]",
                _("patronymic[pre]"),
            ),
            "1y": (
                "_raw_patro_surname_only(raw_data['surname_list'])",
                "patronymic[sur]",
                _("patronymic[sur]"),
            ),
            "2y": (
                "_raw_patro_conn_only(raw_data['surname_list'])",
                "patronymic[con]",
                _("patronymic[con]"),
            ),
            "o": (
                "_raw_nonpatro_surname(raw_data['surname_list'])",
                "notpatronymic",
                _("notpatronymic"),
            ),
            "r": (
                "_raw_nonprimary_surname(raw_data['surname_list'])",
                "rest",
                _("rest", "Remaining names"),
            ),
            "p": (
                "_raw_prefix_surname(raw_data['surname_list'])",
                "prefix",
                _("prefix"),
            ),
            "q": (
                "_raw_single_surname(raw_data['surname_list'])",
                "rawsurnames",
                _("rawsurnames"),
            ),
            "n": ("raw_data['nick']", "nickname", _("nickname")),
            "g": ("raw_data['famnick']", "familynick", _("familynick")),
        }
        args = "raw_data"
        return self._make_fn(format_str, d, args)

    def _gen_cooked_func(self, format_str):
        """The job of building the name from a format string is rather
        expensive and it is called lots and lots of times. So it is worth
        going to some length to optimise it as much as possible.

        This method constructs a new function that is specifically written
        to format a name given a particular format string. This is worthwhile
        because the format string itself rarely changes, so by caching the new
        function and calling it directly when asked to format a name to the
        same format string again we can be as quick as possible.

        The new function is of the form::

        def fn(first, raw_surname_list, suffix, title, call, nick, famnick):
            return "%s %s" % (first,suffix)

        Specific symbols for parts of a name are defined (keywords given):
        't' : title = title
        'f' : given = given (first names)
        'l' : surname = full surname (lastname)
        'c' : call = callname
        'x' : common = nick name, call, or otherwise first first name (common name)
        'i' : initials = initials of the first names
        'm' : primary = primary surname (main)
        '0m': primary[pre]= prefix primary surname (main)
        '1m': primary[sur]= surname primary surname (main)
        '2m': primary[con]= connector primary surname (main)
        'y' : patronymic = pa/matronymic surname (father/mother) - assumed unique
        '0y': patronymic[pre] = prefix      "
        '1y': patronymic[sur] = surname     "
        '2y': patronymic[con] = connector   "
        'o' : notpatronymic = surnames without pa/matronymic and primary
        'r' : rest = non primary surnames
        'p' : prefix = list of all prefixes
        'q' : rawsurnames = surnames without prefixes and connectors
        's' : suffix = suffix
        'n' : nickname = nick name
        'g' : familynick = family nick name

        """

        # we need the names of each of the variables or methods that are
        # called to fill in each format flag.
        # Dictionary is "code": ("expression", "keyword", "i18n-keyword")
        d = {
            "t": ("title", "title", _("title", "Person")),
            "f": ("first", "given", _("given")),
            "l": ("_raw_full_surname(raw_surname_list)", "surname", _("surname")),
            "s": ("suffix", "suffix", _("suffix")),
            "c": ("call", "call", _("call", "Name")),
            "x": (
                "(nick or call or first.split(' ')[0])",
                "common",
                _("common", "Name"),
            ),
            "i": (
                "''.join([word[0] +'.' for word in ('. ' + first).split()][1:])",
                "initials",
                _("initials"),
            ),
            "m": (
                "_raw_primary_surname(raw_surname_list)",
                "primary",
                _("primary", "Name"),
            ),
            "0m": (
                "_raw_primary_prefix_only(raw_surname_list)",
                "primary[pre]",
                _("primary[pre]"),
            ),
            "1m": (
                "_raw_primary_surname_only(raw_surname_list)",
                "primary[sur]",
                _("primary[sur]"),
            ),
            "2m": (
                "_raw_primary_conn_only(raw_surname_list)",
                "primary[con]",
                _("primary[con]"),
            ),
            "y": (
                "_raw_patro_surname(raw_surname_list)",
                "patronymic",
                _("patronymic"),
            ),
            "0y": (
                "_raw_patro_prefix_only(raw_surname_list)",
                "patronymic[pre]",
                _("patronymic[pre]"),
            ),
            "1y": (
                "_raw_patro_surname_only(raw_surname_list)",
                "patronymic[sur]",
                _("patronymic[sur]"),
            ),
            "2y": (
                "_raw_patro_conn_only(raw_surname_list)",
                "patronymic[con]",
                _("patronymic[con]"),
            ),
            "o": (
                "_raw_nonpatro_surname(raw_surname_list)",
                "notpatronymic",
                _("notpatronymic"),
            ),
            "r": (
                "_raw_nonprimary_surname(raw_surname_list)",
                "rest",
                _("rest", "Remaining names"),
            ),
            "p": ("_raw_prefix_surname(raw_surname_list)", "prefix", _("prefix")),
            "q": (
                "_raw_single_surname(raw_surname_list)",
                "rawsurnames",
                _("rawsurnames"),
            ),
            "n": ("nick", "nickname", _("nickname")),
            "g": ("famnick", "familynick", _("familynick")),
        }
        args = "first,raw_surname_list,suffix,title,call,nick,famnick"
        return self._make_fn(format_str, d, args)

    def format_str(self, name, format_str):
        return self._format_str_base(
            name.first_name,
            name.surname_list,
            name.suffix,
            name.title,
            name.call,
            name.nick,
            name.famnick,
            format_str,
        )

    def format_str_raw(self, raw_data, format_str):
        """
        Format a name from the raw name list. To make this as fast as possible
        this uses :func:`_gen_raw_func` to generate a new method for each new
        format_string.

        Is does not call :meth:`_format_str_base` because it would introduce an
        extra method call and we need all the speed we can squeeze out of this.
        """
        func = self.__class__.raw_format_funcs.get(format_str)
        if func is None:
            func = self._gen_raw_func(format_str)
            self.__class__.raw_format_funcs[format_str] = func

        return func(raw_data)

    def _format_str_base(
        self, first, surname_list, suffix, title, call, nick, famnick, format_str
    ):
        """
        Generates name from a format string.

        The following substitutions are made:
        '%t' : title
        '%f' : given (first names)
        '%l' : full surname (lastname)
        '%c' : callname
        '%x' : nick name, call, or otherwise first first name (common name)
        '%i' : initials of the first names
        '%m' : primary surname (main)
        '%0m': prefix primary surname (main)
        '%1m': surname primary surname (main)
        '%2m': connector primary surname (main)
        '%y' : pa/matronymic surname (father/mother) - assumed unique
        '%0y': prefix      "
        '%1y': surname     "
        '%2y': connector   "
        '%o' : surnames without patronymic
        '%r' : non-primary surnames (rest)
        '%p' : list of all prefixes
        '%q' : surnames without prefixes and connectors
        '%s' : suffix
        '%n' : nick name
        '%g' : family nick name
        The capital letters are substituted for capitalized name components.
        The %% is substituted with the single % character.
        All the other characters in the fmt_str are unaffected.
        """
        func = self.__class__.format_funcs.get(format_str)
        if func is None:
            func = self._gen_cooked_func(format_str)
            self.__class__.format_funcs[format_str] = func
        try:
            s = func(
                first,
                [object_to_data(surn) for surn in surname_list],
                suffix,
                title,
                call,
                nick,
                famnick,
            )
        except (
            ValueError,
            TypeError,
        ):
            raise NameDisplayError("Incomplete format string")

        return s

    # -------------------------------------------------------------------------

    def primary_surname(self, name):
        global PAT_AS_SURN
        nrsur = len(name.surname_list)
        sur = name.get_primary_surname()
        if (
            not PAT_AS_SURN
            and nrsur <= 1
            and (
                sur.get_origintype().value == _ORIGINPATRO
                or sur.get_origintype().value == _ORIGINMATRO
            )
        ):
            return ""
        return sur.get_surname()

    def sort_string(self, name):
        return "%-25s%-30s%s" % (
            self.primary_surname(name),
            name.first_name,
            name.suffix,
        )

    def sorted(self, person):
        """
        Return a text string representing the :class:`~.person.Person`
        instance's :class:`~.name.Name` in a manner that should be used for
        displaying a sortedname.

        :param person: :class:`~.person.Person` instance that contains the
                       :class:`~.name.Name` that is to be displayed. The
                       primary name is used for the display.
        :type person: :class:`~.person.Person`
        :returns: Returns the :class:`~.person.Person` instance's name
        :rtype: str
        """
        name = person.get_primary_name()
        return self.sorted_name(name)

    def sorted_name(self, name):
        """
        Return a text string representing the :class:`~.name.Name` instance
        in a manner that should be used for sorting the name in a list.

        :param name: :class:`~.name.Name` instance that is to be displayed.
        :type name: :class:`~.name.Name`
        :returns: Returns the :class:`~.name.Name` string representation
        :rtype: str
        """
        num = self._is_format_valid(name.sort_as)
        return self.name_formats[num][_F_FN](name)

    def truncate(self, full_name, max_length=15, elipsis="..."):
        name_out = ""
        if len(full_name) <= max_length:
            name_out = full_name
        else:
            last_space = full_name.rfind(" ", max_length)
            if (last_space) > -1:
                name_out = full_name[:last_space]
            else:
                name_out = full_name[:max_length]
            name_out += " " + elipsis
        return name_out

    def raw_sorted_name(self, raw_data):
        """
        Return a text string representing the :class:`~.name.Name` instance
        in a manner that should be used for sorting the name in a list.

        :param name: raw unserialized data of name that is to be displayed.
        :type name: tuple
        :returns: Returns the :class:`~.name.Name` string representation
        :rtype: str
        """
        num = self._is_format_valid(raw_data["sort_as"])
        return self.name_formats[num][_F_RAWFN](raw_data)

    def display(self, person):
        """
        Return a text string representing the :class:`~.person.Person`
        instance's :class:`~.name.Name` in a manner that should be used for
        normal displaying.

        :param person: :class:`~.person.Person` instance that contains the
                       :class:`~.name.Name` that is to be displayed. The
                       primary name is used for the display.
        :type person: :class:`~.person.Person`
        :returns: Returns the :class:`~.person.Person` instance's name
        :rtype: str
        """
        name = person.get_primary_name()
        return self.display_name(name)

    def display_format(self, person, num):
        """
        Return a text string representing the L{gen.lib.Person} instance's
        L{Name} using num format.

        @param person: L{gen.lib.Person} instance that contains the
        L{Name} that is to be displayed. The primary name is used for
        the display.
        @type person: L{gen.lib.Person}
        @param num: num of the format to be used, as return by
        name_displayer.add_name_format('name','format')
        @type num: int
        @returns: Returns the L{gen.lib.Person} instance's name
        @rtype: str
        """
        name = person.get_primary_name()
        return self.name_formats[num][_F_FN](name)

    def display_formal(self, person):
        """
        Return a text string representing the :class:`~.person.Person`
        instance's :class:`~.name.Name` in a manner that should be used for
        formal displaying.

        :param person: :class:`~.person.Person` instance that contains the
                       :class:`~.name.Name` that is to be displayed. The
                       primary name is used for the display.
        :type person: :class:`~.person.Person`
        :returns: Returns the :class:`~.person.Person` instance's name
        :rtype: str
        """
        # FIXME: At this time, this is just duplicating display() method
        name = person.get_primary_name()
        return self.display_name(name)

    def display_name(self, name):
        """
        Return a text string representing the :class:`~.name.Name` instance
        in a manner that should be used for normal displaying.

        :param name: :class:`~.name.Name` instance that is to be displayed.
        :type name: :class:`~.name.Name`
        :returns: Returns the :class:`~.name.Name` string representation
        :rtype: str
        """
        if name is None:
            return ""

        num = self._is_format_valid(name.display_as)
        return self.name_formats[num][_F_FN](name)

    def raw_display_name(self, raw_data):
        """
        Return a text string representing the :class:`~.name.Name` instance
        in a manner that should be used for normal displaying.

        :param name: raw unserialized data of name that is to be displayed.
        :type name: tuple
        :returns: Returns the :class:`~.name.Name` string representation
        :rtype: str
        """
        num = self._is_format_valid(raw_data["display_as"])
        return self.name_formats[num][_F_RAWFN](raw_data)

    def display_given(self, person):
        return self.format_str(person.get_primary_name(), "%f")

    def name_grouping(self, db, person):
        """
        Return the name under which to group this person. This is defined as:

        1. if group name is defined on primary name, use that
        2. if group name is defined for the primary surname of the primary
           name, use that
        3. use primary surname of primary name otherwise
        """
        return self.name_grouping_name(db, person.primary_name)

    def name_grouping_name(self, db, pn):
        """
        Return the name under which to group. This is defined as:

        1. if group name is defined, use that
        2. if group name is defined for the primary surname, use that
        3. use primary surname itself otherwise

        :param pn: :class:`~.name.Name` object
        :type pn: :class:`~.name.Name` instance
        :returns: Returns the groupname string representation
        :rtype: str
        """
        if pn.group_as:
            return pn.group_as
        return db.get_name_group_mapping(pn.get_primary_surname().get_surname())

    def name_grouping_data(self, db, pn):
        """
        Return the name under which to group. This is defined as:

        1. if group name is defined, use that
        2. if group name is defined for the primary surname, use that
        3. use primary surname itself otherwise
        4. if no primary surname, do we have a ma/patronymic surname ?
           in this case, group name will be the ma/patronymic name.

        :param pn: raw unserialized data of name
        :type pn: tuple
        :returns: Returns the groupname string representation
        :rtype: str
        """
        if pn["group_as"]:
            return pn["group_as"]
        name = pn["group_as"]
        if not name:
            # if we have no primary surname, perhaps we have a
            # patronymic/matronynic name ?
            srnme = pn["surname_list"]
            surname = []
            for _surname in srnme:
                if (
                    _surname["origintype"]["value"] == _ORIGINPATRO
                    or _surname["origintype"]["value"] == _ORIGINMATRO
                ):
                    # Yes, we have one.
                    surname = [_surname]
            # name1 is the ma/patronymic name.
            name1 = _raw_patro_surname_only(surname)
            if name1 and len(srnme) == 1:
                name = db.get_name_group_mapping(name1)
        if not name:
            name = db.get_name_group_mapping(
                _raw_primary_surname_only(pn["surname_list"])
            )
        return name

    def _make_fn(self, format_str, d, args):
        """
        Create the name display function and handles dependent punctuation.
        """
        # d is a dict: dict[code] = (expr, word, translated word)

        # First, go through and do internationalization-based
        # key-word replacement. Just replace ikeywords with
        # %codes (ie, replace "irstnamefay" with "%f", and
        # "IRSTNAMEFAY" for %F)

        if len(format_str) > 2 and format_str[0] == format_str[-1] == '"':
            pass
        else:
            d_keys = [(code, _tuple[2]) for code, _tuple in d.items()]
            d_keys.sort(
                key=_make_cmp_key, reverse=True
            )  # reverse on length and by ikeyword
            for code, ikeyword in d_keys:
                exp, keyword, ikeyword = d[code]
                format_str = format_str.replace(ikeyword, "%" + code)
                format_str = format_str.replace(ikeyword.title(), "%" + code)
                format_str = format_str.replace(ikeyword.upper(), "%" + code.upper())
        # Next, go through and do key-word replacement.
        # Just replace keywords with
        # %codes (ie, replace "firstname" with "%f", and
        # "FIRSTNAME" for %F)
        if len(format_str) > 2 and format_str[0] == format_str[-1] == '"':
            pass
        else:
            d_keys = [(code, _tuple[1]) for code, _tuple in d.items()]
            d_keys.sort(
                key=_make_cmp_key, reverse=True
            )  # reverse sort on length and by keyword
            # if in double quotes, just use % codes
            for code, keyword in d_keys:
                exp, keyword, ikeyword = d[code]
                format_str = format_str.replace(keyword, "%" + code)
                format_str = format_str.replace(keyword.title(), "%" + code)
                format_str = format_str.replace(keyword.upper(), "%" + code.upper())
        # Get lower and upper versions of codes:
        codes = list(d.keys()) + [c.upper() for c in d]
        # Next, list out the matching patterns:
        # If it starts with "!" however, treat the punctuation verbatim:
        if len(format_str) > 0 and format_str[0] == "!":
            patterns = [
                "%(" + ("|".join(codes)) + ")",  # %s
            ]
            format_str = format_str[1:]
        else:
            patterns = [
                ',\\W*"%(' + ("|".join(codes)) + ')"',  # ,\W*"%s"
                ",\\W*\\(%(" + ("|".join(codes)) + ")\\)",  # ,\W*(%s)
                ",\\W*%(" + ("|".join(codes)) + ")",  # ,\W*%s
                '"%(' + ("|".join(codes)) + ')"',  # "%s"
                "_%(" + ("|".join(codes)) + ")_",  # _%s_
                "\\(%(" + ("|".join(codes)) + ")\\)",  # (%s)
                "%(" + ("|".join(codes)) + ")",  # %s
            ]
        new_fmt = format_str

        # replace the specific format string flags with a
        # flag that works in standard python format strings.
        new_fmt = re.sub("|".join(patterns), "%s", new_fmt)

        # replace special meaning codes we need to have verbatim in output
        if len(new_fmt) > 2 and new_fmt[0] == new_fmt[-1] == '"':
            new_fmt = new_fmt.replace("\\", r"\\")
            new_fmt = new_fmt[1:-1].replace('"', r"\"")
        else:
            new_fmt = new_fmt.replace("\\", r"\\")
            new_fmt = new_fmt.replace('"', '\\"')

        # find each format flag in the original format string
        # for each one we find the variable name that is needed to
        # replace it and add this to a list. This list will be used to
        # generate the replacement tuple.

        # This compiled pattern should match all of the format codes.
        pat = re.compile("|".join(patterns))
        param = ()
        mat = pat.search(format_str)
        while mat:
            match_pattern = mat.group(0)  # the matching pattern
            # prefix, code, suffix:
            p, code, s = re.split("%(.)", match_pattern)
            if code in "0123456789":
                code = code + s[0]
                s = s[1:]
            field = d[code.lower()][0]
            if code.isupper():
                field += ".upper()"
            if p == "" and s == "":
                param = param + (field,)
            else:
                param = param + ("ifNotEmpty(%s,'%s','%s')" % (field, p, s),)
            mat = pat.search(format_str, mat.end())
        s = """
def fn(%s):
    def ifNotEmpty(str,p,s):
        if str == '':
            return ''
        else:
            return p + str + s
    return cleanup_name("%s" %% (%s))""" % (
            args,
            new_fmt,
            ",".join(param),
        )
        try:
            result = {}
            exec(s, globals(), result)
            return result["fn"]
        except SyntaxError:
            LOG.error(
                "Wrong name format string %s\n"
                "ERROR, Edit Name format in Preferences->Display to correct\n"
                + _("Wrong name format string %s")
                + "\nERROR, Edit Name format in Preferences->Display to correct",
                new_fmt,
                new_fmt,
            )

            def errfn(*arg):
                return _("ERROR, Edit Name format in Preferences")

            return errfn


displayer = NameDisplay()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
# Copyright (C) 2010       Michiel D. Nauta
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
LdsOrdBase class for Gramps.
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .const import EQUAL, IDENTICAL
from .ldsord import LdsOrd


# -------------------------------------------------------------------------
#
# LdsOrdBase
#
# -------------------------------------------------------------------------
class LdsOrdBase:
    """
    Base class for lds_ord-aware objects.
    """

    def __init__(self, source=None):
        """
        Initialize a LdsOrdBase.

        If the source is not None, then object is initialized from values of
        the source object.

        :param source: Object used to initialize the new object
        :type source: LdsOrdBase
        """

        if source:
            self.lds_ord_list = [LdsOrd(lds_ord) for lds_ord in source.lds_ord_list]
        else:
            self.lds_ord_list = []

    def serialize(self):
        """
        Convert the object to a serialized tuple of data.
        """
        return [lds_ord.serialize() for lds_ord in self.lds_ord_list]

    def unserialize(self, data):
        """
        Convert a serialized tuple of data to an object
        """
        self.lds_ord_list = [LdsOrd().unserialize(item) for item in data]

    def add_lds_ord(self, lds_ord):
        """
        Add the :class:`~.ldsord.LdsOrd` instance to the object's list of
        lds_ords.

        :param lds_ord: :class:`~.ldsord.LdsOrd` instance to add to the object's
                        lds_ord list
        :type lds_ord: list
        """
        self.lds_ord_list.append(lds_ord)

    def remove_lds_ord(self, lds_ord):
        """
        Remove the specified :class:`~.ldsord.LdsOrd` instance from the lds_ord
        list.

        If the instance does not exist in the list, the operation has no effect.

        :param lds_ord: :class:`~.ldsord.LdsOrd` instance to remove from the
                        list
        :type lds_ord: :class:`~.ldsord.LdsOrd`

        :returns: True if the lds_ord was removed, False if it was not in the
                  list.
        :rtype: bool
        """
        if lds_ord in self.lds_ord_list:
            self.lds_ord_list.remove(lds_ord)
            return True
        return False

    def get_lds_ord_list(self):
        """
        Return the list of :class:`~.ldsord.LdsOrd` instances associated with
        the object.

        :returns: Returns the list of :class:`~.ldsord.LdsOrd` instances
        :rtype: list
        """
        return self.lds_ord_list

    def set_lds_ord_list(self, lds_ord_list):
        """
        Assign the passed list to the object's list of :class:`~.ldsord.LdsOrd`
        instances.

        :param lds_ord_list: List of :class:`~.ldsord.LdsOrd` instances to be
                             associated with the object
        :type lds_ord_list: list
        """
        self.lds_ord_list = lds_ord_list

    def _merge_lds_ord_list(self, acquisition):
        """
        Merge the list of ldsord from acquisition with our own.

        :param acquisition: the ldsord list of this object will be merged with
                            the current ldsord list.
        :type acquisition: LdsOrdBase
        """
        ldsord_list = self.lds_ord_list[:]
        for addendum in acquisition.get_lds_ord_list():
            for ldsord in ldsord_list:
                equi = ldsord.is_equivalent(addendum)
                if equi == IDENTICAL:
                    break
                if equi == EQUAL:
                    ldsord.merge(addendum)
                    break
            else:
                self.lds_ord_list.append(addendum)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
# Copyright (C) 2009       Gary Burton
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Provide the managed window interface, which allows Gramps to track
the create/deletion of dialog windows.
"""

# -------------------------------------------------------------------------
#
# Standard python modules
#
# -------------------------------------------------------------------------
import os
from io import StringIO
import html
import logging

# -------------------------------------------------------------------------
#
# Set up logging
#
# -------------------------------------------------------------------------
_LOG = logging.getLogger(".")
# -------------------------------------------------------------------------
#
# GNOME/GTK
#
# -------------------------------------------------------------------------
from gi.repository import Gtk
from gi.repository import Gdk

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GLADE_FILE, ICON
from gramps.gen.errors import WindowActiveError
from gramps.gen.config import config
from gramps.gen.constfunc import is_quartz
from .uimanager import ActionGroup, valid_action_name
from .utils import get_display_size
from .glade import Glade

# -------------------------------------------------------------------------
#
# Window manager
#
# -------------------------------------------------------------------------

_win_top = '<section id="WinMenu">\n'
_win_btm = "</section>\n"
DISABLED = -1

# -----------------------------------------------------------------------
#
# Helper function
#
# -----------------------------------------------------------------------


def get_object(self, value):
    raise DeprecationWarning("ManagedWindow.get_object: shouldn't get here")
    if self.get_name() == value:
        return self
    elif hasattr(self, "get_children"):
        for child in self.get_children():
            object = get_object(child, value)
            if object is not None:
                return object
    return None


class GrampsWindowManager:
    """
    Manage hierarchy of open Gramps windows.

    This class's purpose is to manage the hierarchy of open windows.
    The idea is to maintain the tree of branches and leaves.
    A leaf does not have children and corresponds to a single open window.
    A branch has children and corresponds to a group of windows.

    We will follow the convention of having first leaf in any given
    branch represent a parent window of the group, and the rest of the
    children leaves/branches represent windows spawned from the parent.

    The tree structure is maintained as a list of items.
    Items which are lists are branches.
    Items which are not lists are leaves.

    Lookup of an item is done via track sequence. The elements of
    the track sequence specify the lookup order: [2,3,1] means
    'take the second item of the tree, take its third child, and
    then the first child of that child'.

    Lookup can be also done by ID for windows that are identifiable.
    """

    def __init__(self, uimanager):
        # initialize empty tree and lookup dictionary
        self.uimanager = uimanager
        self.window_tree = []
        self.id2item = {}
        self.action_group = ActionGroup(name="WindowManger")
        self.active = DISABLED
        self.ui = _win_top + _win_btm

    def disable(self):
        """
        Remove the UI and action groups if the navigation is enabled
        """
        if self.active != DISABLED:
            self.uimanager.remove_ui(self.active)
            self.uimanager.remove_action_group(self.action_group)
            self.active = DISABLED

    def enable(self):
        """
        Enables the UI and action groups
        """
        self.uimanager.insert_action_group(self.action_group)
        self.active = self.uimanager.add_ui_from_string([self.ui])
        self.uimanager.update_menu()

    def get_item_from_track(self, track):
        # Recursively find an item given track sequence
        item = self.window_tree
        for index in track:
            item = item[index]
        return item

    def get_item_from_id(self, item_id):
        # Find an item given its ID
        # Return None if the ID is not found
        return self.id2item.get(item_id, None)

    def get_item_from_window(self, window):
        """This finds a ManagedWindow from a Gtk top_level object (typicaly
        a window).

        For example, to find my managedwindow track within a class of Gtk
        widget:
            mywindow = self.get_toplevel()  # finds top level Gtk object
            managed_window = self.uistate.gwm.get_item_from_window(mywindow)
            track = managed_window.track
        """
        for key, item in self.id2item.items():
            if item.window == window:
                return self.id2item[key]
        return None

    def find_modal_window(self, window):
        """This finds a ManagedWindow that is modal, if any, excluding the
        'window' that is a parameter.  There should be only one.
        If no ManagedWindow is modal, returns None.
        """
        for dummy, item in self.id2item.items():
            if item.window != window and item.window.get_modal():
                return item.window
        return None

    def close_track(self, track):
        # This is called when item needs to be closed
        # Closes all its children and then removes the item from the tree.
        try:
            item = self.get_item_from_track(track)
            self.recursive_action(item, self.close_item)
            # This only needs to be run once for the highest level point
            # to remove.
            self.remove_item(track)
        except IndexError:
            print("Missing item from window manager", track, self.close_item)

    def recursive_action(self, item, func, *args):
        # This function recursively calls itself over the child items
        # starting with the given item.
        # Eventualy, every non-list item (leaf) will be reached
        # and the func(item,*args) will be called on that item.
        if isinstance(item, list):
            # If this item is a branch
            # close the children except for the first one
            for sub_item in item[1:]:
                self.recursive_action(sub_item, func, *args)
            # return the first child
            last_item = item[0]
        else:
            # This item is a leaf -- no children to close
            # return itself
            last_item = item
        func(last_item, *args)

    def close_item(self, item, *args):
        # Given an item, close its window and remove it's ID from the dict
        if item.opened:
            item.close()
        if item.window_id:
            del self.id2item[item.window_id]
            item.window_id = None
        if item.get_window():
            item.get_window().destroy()
            item.window = None

    def remove_item(self, track):
        # We need the whole gymnastics below because our item
        # may actually be a list consisting of a single real
        # item and empty lists.

        # find the track corresponding to the parent item
        parent_track = track[:-1]
        # find index of our item in parent
        child_in_parent = track[-1:][0]
        # obtain parent item and remove our item from it
        parent_item = self.get_item_from_track(parent_track)
        parent_item.pop(child_in_parent)
        # Adjust each item following the removed one
        # so that it's track is down by one on this level
        for ix in range(child_in_parent, len(parent_item)):
            item = parent_item[ix]
            self.recursive_action(item, self.move_item_down, len(track) - 1)
        # Rebuild menu
        self.build_windows_menu()

    def move_item_down(self, item, *args):
        # Given an item and an index, adjust the item's track
        # by subtracting 1 from that index's level
        index = args[0]
        item.track[index] -= 1

    def add_item(self, track, item):
        # if the item is identifiable then we need to remember
        # its id so that in the future we recall this window
        # instead of spawning a new one
        # So people can make as many windows of type item.window_id = None
        # Use this for add dialogs, where users may add as many values
        # simultaneously as they want.
        # Actually, we should do this away, as requiring at least id(obj)
        #  is not a big requirement ?
        if item.window_id:
            self.id2item[item.window_id] = item

        # Make sure we have a track
        parent_item = self.get_item_from_track(track)
        assert (
            isinstance(parent_item, list) or track == []
        ), "Gwm: add_item: Incorrect track - Is parent not a leaf?"

        # Prepare a new item, depending on whether it is branch or leaf
        if item.submenu_label:
            # This is an item with potential children -- branch
            new_item = [item]
        else:
            # This is an item without children -- leaf
            new_item = item

        # append new item to the parent
        parent_item.append(new_item)

        # rebuild the Windows menu based on the new tree
        self.build_windows_menu()

        # prepare new track corresponding to the added item and return it
        new_track = track + [len(parent_item) - 1]
        return new_track

    def call_back_factory(self, item):
        if not isinstance(item, list):

            def func(*obj):
                if item.window_id and self.id2item.get(item.window_id):
                    self.id2item[item.window_id]._present()

        else:

            def func(*obj):
                pass

        return func

    def generate_id(self, item):
        return valid_action_name("wm-" + str(item.window_id))

    def display_menu_list(self, data, action_data, mlist):
        menuitem = (
            "<item>\n"
            '<attribute name="action">win.%s</attribute>\n'
            '<attribute name="label" translatable="yes">'
            "%s...</attribute>\n"
            "</item>\n"
        )
        if isinstance(mlist, (list, tuple)):
            i = mlist[0]
            idval = self.generate_id(i)
            data.write(
                '<submenu>\n<attribute name="label"'
                ' translatable="yes">%s</attribute>\n' % html.escape(i.submenu_label)
            )
        else:
            i = mlist
            idval = self.generate_id(i)

        data.write(menuitem % (idval, html.escape(i.menu_label)))
        action_data.append((idval, self.call_back_factory(i)))

        if isinstance(mlist, (list, tuple)) and (len(mlist) > 1):
            for i in mlist[1:]:
                if isinstance(i, list):
                    self.display_menu_list(data, action_data, i)
                else:
                    idval = self.generate_id(i)
                    data.write(menuitem % (idval, html.escape(i.menu_label)))
                    action_data.append((idval, self.call_back_factory(i)))
        if isinstance(mlist, (list, tuple)):
            data.write("</submenu>\n")

    def build_windows_menu(self):
        if self.active != DISABLED:
            self.uimanager.remove_ui(self.active)
            self.uimanager.remove_action_group(self.action_group)

        self.action_group = ActionGroup(name="WindowManger")
        action_data = []

        data = StringIO()
        data.write(_win_top)
        for i in self.window_tree:
            self.display_menu_list(data, action_data, i)
        data.write(_win_btm)
        self.ui = data.getvalue()
        data.close()
        self.action_group.add_actions(action_data)
        self.enable()


# -------------------------------------------------------------------------
#
# Gramps Managed Window class
#
# -------------------------------------------------------------------------
class ManagedWindow:
    """
    Managed window base class.

    This class provides all the goodies necessary for user-friendly window
    management in GRAMPS: registering the menu item under the Windows
    menu, keeping track of child windows, closing them on close/delete
    event, and presenting itself when selected or attempted to create again.
    """

    def __init__(self, uistate, track, obj, modal=False):
        """
        Create child windows and add itself to menu, if not there already.


        The usage from derived classes is envisioned as follows:


        from .managedwindow import ManagedWindow
        class SomeWindowClass(ManagedWindow):
            def __init__(self, uistate, track, obj, modal):
                window_id = self        # Or e.g. window_id = person.handle
                ManagedWindow.__init__(self,
                                       uistate,
                                       track,
                                       window_id,
                                       modal=False)
                # Proceed with the class.
                window = Gtk.Dialog()  # Some Gtk window object to manage
                self.set_window(window, None, None)  # See set_window def below
                # setup window size, position tracking configuration
                self.setup_configs(self, "interface.mywindow", 680, 400)
                ...
                self.close()

            def build_menu_names(self, obj):
                ''' Define menu labels.  If your ManagedWindow can have
                ManagedWindow children, you must include a submenu_label
                string; However, if modal, that string is never seen and can
                be ' '.
                '''
                submenu_label = None    # This window cannot have children
                menu_label = 'Menu label for this window'
                return (menu_label, submenu_label)

        :param uistate:  gramps uistate
        :param track:    {list of parent windows, [] if the main Gramps window
                            is the parent}
        :param obj:      The object that is used to id the managed window,
                            The inheriting object needs a method
                            build_menu_names(self, obj)
                            which works on this obj and creates menu labels
                            for use in the Gramps Window Menu.
                            If self.submenu_label ='' then leaf, else branch
        :param modal:    True/False, if True, this window is made modal
                            (always on top, and always has focus).  Any child
                            windows are also automatically made modal by moving
                            the modal flag to the child.  On close of a child
                            the modal flag is sent back to the parent.

        If a modal window is used and has children, its and any child 'track'
        parameters must properly be set to the parents 'self.track'.
        Only one modal window can be supported by Gtk without potentially
        hiding of a modal window while it has focus.  So don't use
        non-managed Gtk windows as children and set them modal.

        If you use the 'Gtk.Dialog.run()' within your ManagedWindow, Gtk makes
        the dialog modal.  Accordingly you MUST also set the ManagedWindow
        modal for proper operation of any child windows.

        You must use 'self.show()' in order for your ManagedWindow to work
        properly, this in turn calls the Gtk.Window.show_all() for you.

        The ManagedWindow uses 'track' to properly do a
        Gtk.Window.set_transient_for() for you; you don't have to do it
        yourself.  If you need a 'parent=' to call an unmanaged window,
        self.window is available.

        """
        window_key = self.build_window_key(obj)
        menu_label, submenu_label = self.build_menu_names(obj)
        self._gladeobj = None
        self.isWindow = None
        self.width_key = None
        self.height_key = None
        self.horiz_position_key = None
        self.vert_position_key = None
        self.__refs_for_deletion = []
        self.modal = modal
        self.other_modal_window = None

        if uistate and uistate.gwm.get_item_from_id(window_key):
            uistate.gwm.get_item_from_id(window_key)._present()
            raise WindowActiveError("This window is already active")
        else:
            self.window_id = window_key
            self.submenu_label = submenu_label
            self.menu_label = menu_label
            self.uistate = uistate
            if uistate:
                self.track = self.uistate.gwm.add_item(track, self)
            else:
                self.track = []
            # Work out parent_window
            if len(self.track) > 1:
                # We don't belong to the lop level
                if self.track[-1] > 0:
                    # If we are not the first in the group,
                    # then first in that same group is our parent
                    parent_item_track = self.track[:-1]
                    parent_item_track.append(0)
                else:
                    # If we're first in the group, then our parent
                    # is the first in the group one level up
                    parent_item_track = self.track[:-2]
                    parent_item_track.append(0)

                # Based on the track, get item and then window object
                managed_parent = self.uistate.gwm.get_item_from_track(parent_item_track)
                self.parent_window = managed_parent.window
            else:
                # On the top level: we use gramps top window
                if self.uistate:
                    self.parent_window = self.uistate.window
                else:
                    self.parent_window = None

    def set_window(self, window, title, text, msg=None, isWindow=False):
        """
        Set the window that is managed.

        :param window:   if isWindow=False window must be a Gtk.Window() object
                         (or a subclass such as Gtk.Dialog), otherwise None
        :param title:    a label widget in which to write the title,
                         else None if not needed
        :param text:     text to use as title of window and in title param
                         can also be None if Glade defines title
        :param msg:      if not None, use msg as title of window instead of text
        :param isWindow: {if isWindow than self is the window
                            (so self inherits from Gtk.Window and
                            from ManagedWindow)
                         if not isWindow, than window is the Window to manage,
                            and after this method self.window stores it.
                        }

        """
        self.isWindow = isWindow
        self.msg = msg
        self.titlelabel = title
        if self.isWindow:
            set_titles(self, title, text, msg)
            self.window = self
        else:
            set_titles(window, title, text, msg)
            # closing the Gtk.Window must also close ManagedWindow
            self.window = window
            self.window.connect("delete-event", self.close)
        # Set the mnemonic modifier on Macs to alt-ctrl so that it
        # doesn't interfere with the extended keyboard, see
        # https://gramps-project.org/bugs/view.php?id=6943
        if is_quartz():
            self.window.set_mnemonic_modifier(
                Gdk.ModifierType.CONTROL_MASK | Gdk.ModifierType.MOD1_MASK
            )

        if self.modal:
            self.window.set_modal(True)
        # The following makes sure that we only have one modal window open;
        # if more the older ones get temporarily made non-modal.
        if self.uistate:
            self.other_modal_window = self.uistate.gwm.find_modal_window(window)
        if self.other_modal_window:
            self.other_modal_window.set_modal(False)
            self.window.set_modal(True)
            self.modal = True

    def get_window(self):
        """
        Return the managed window.
        """
        if self.isWindow:
            return self
        else:
            return self.window

    def update_title(self, text):
        if self.isWindow:
            set_titles(self, self.titlelabel, text, self.msg)
        else:
            set_titles(self.window, self.titlelabel, text, self.msg)

    def build_menu_names(self, obj):
        return ("Undefined Menu", "Undefined Submenu")

    def build_window_key(self, obj):
        return id(obj)

    def define_glade(self, top_module, glade_file=None, also_load=[]):
        if glade_file is None:
            raise TypeError("ManagedWindow.define_glade: no glade file")
            glade_file = GLADE_FILE
        self._gladeobj = Glade(glade_file, None, top_module, also_load)
        return self._gladeobj

    def get_widget(self, name):
        assert self._gladeobj
        object = self._gladeobj.get_child_object(name)
        if object is not None:
            return object
        raise ValueError(
            'ManagedWindow.get_widget: "%s" widget not found in "%s/%s"'
            % (name, self._gladeobj.dirname, self._gladeobj.filename)
        )
        return object

    def connect_button(self, button_name, function):
        assert self._gladeobj
        self.get_widget(button_name).connect("clicked", function)

    def show(self):
        """The following covers a case where there are multiple modal windows
        to be open; possibly not in parent child relation.  If this occurs,
        we use most recent modal window as parent.  This occurs during startup
        when both the 'Available Gramps Updates for Addons' and 'Family Trees'
        windows are started by the viewmanager.
        """
        assert self.window, "ManagedWindow: self.window does not exist!"
        if self.other_modal_window:
            self.window.set_transient_for(self.other_modal_window)
        else:
            self.window.set_transient_for(self.parent_window)
        self.opened = True
        self.window.show_all()

    def close(self, *obj):
        """
        Close itself.

        Takes care of closing children and removing itself from menu.
        """
        if hasattr(self, "opened") and not self.opened:
            _LOG.warning("Tried to close a ManagedWindow more than once.")
            return  # in case close somehow gets called again
        self.opened = False
        self._save_position(save_config=False)  # the next line will save it
        self._save_size()
        self.uistate.gwm.close_track(self.track)
        self.clean_up()
        # put a previously modal window back to modal, now that we are closing
        if self.other_modal_window:
            self.other_modal_window.set_modal(True)
        self.parent_window.present()

    def _present(self):
        """
        Present window (unroll/unminimize/bring to top).
        """
        assert hasattr(self, "window"), "ManagedWindow: self.window does not exist!"
        self.window.present()

    def _set_size(self):
        """
        Set the dimensions of the window
        """
        # self.width_key is set in the subclass (or in setup_configs)
        if self.width_key is not None:
            width = config.get(self.width_key)
            height = config.get(self.height_key)
            self.window.resize(width, height)

    def _save_size(self):
        """
        Save the dimensions of the window to the config file
        """
        # self.width_key is set in the subclass (or in setup_configs)
        if self.width_key is not None:
            (width, height) = self.window.get_size()
            config.set(self.width_key, width)
            config.set(self.height_key, height)
            config.save()

    def _set_position(self):
        """
        Set the position of the window
        """
        # self.horiz_position_key is set in the subclass (or in setup_configs)
        if self.horiz_position_key is not None:
            horiz_position = config.get(self.horiz_position_key)
            vert_position = config.get(self.vert_position_key)
            # make sure some of left side shows on screen
            # for part time multi monitor setups
            s_width, s_height = get_display_size(self.window)
            if horiz_position > s_width - 50 or vert_position > s_height - 50:
                (p_width, p_height) = self.parent_window.get_size()
                (p_horiz, p_vert) = self.parent_window.get_position()
                (w_width, w_height) = self.window.get_size()
                horiz_position = p_horiz + ((p_width - w_width) // 2)
                vert_position = p_vert + ((p_height - w_height) // 2)
            self.window.move(horiz_position, vert_position)

    def _save_position(self, save_config=True):
        """
        Save the window's position to the config file

        (You can set save_config False if a _save_size() will instantly follow)
        """
        # self.horiz_position_key is set in the subclass (or in setup_configs)
        if self.horiz_position_key is not None:
            (horiz_position, vert_position) = self.window.get_position()
            config.set(self.horiz_position_key, horiz_position)
            config.set(self.vert_position_key, vert_position)
            if save_config:
                config.save()

    def setup_configs(
        self,
        config_base,
        default_width,
        default_height,
        default_horiz_position=None,
        default_vert_position=None,
        p_width=None,
        p_height=None,
    ):  # for fullscreen
        """
        Helper method to setup the window's configuration settings

        @param config_base: the common config name, e.g. 'interface.clipboard'
        @type config_base: str
        @param default_width, default_height: the default width and height
        @type default_width, default_height: int
        @param default_horiz_position, default_vert_position: if either is None
            then that position is centered on the parent, else explicitly set
        @type default_horiz_position, default_vert_position: int or None
        @param p_width, p_height: the parent's width and height
        @type p_width, p_height: int or None
        """
        self.width_key = config_base + "-width"
        self.height_key = config_base + "-height"
        self.horiz_position_key = config_base + "-horiz-position"
        self.vert_position_key = config_base + "-vert-position"
        if p_width is None and p_height is None:  # default case
            (p_width, p_height) = self.parent_window.get_size()
            (p_horiz, p_vert) = self.parent_window.get_position()
        else:
            p_horiz = p_vert = 0  # fullscreen
        if default_horiz_position is None:
            default_horiz_position = p_horiz + ((p_width - default_width) // 2)
        if default_vert_position is None:
            default_vert_position = p_vert + ((p_height - default_height) // 2)
        config.register(self.width_key, default_width)
        config.register(self.height_key, default_height)
        config.register(self.horiz_position_key, default_horiz_position)
        config.register(self.vert_position_key, default_vert_position)
        self._set_size()
        self._set_position()

    def track_ref_for_deletion(self, ref):
        """
        Record references of instance variables that need to be removed
        from scope so that the class can be garbage collected
        """
        if ref not in self.__refs_for_deletion:
            self.__refs_for_deletion.append(ref)

    def clean_up(self):
        """
        Remove any instance variables from scope which point to non-glade
        GTK objects so that the class can be garbage collected.
        If the object is a Gramps widget then it should have a clean_up method
        which can be called that removes any other GTK object it contains.
        """
        while len(self.__refs_for_deletion):
            attr = self.__refs_for_deletion.pop()
            obj = getattr(self, attr)
            if hasattr(obj, "clean_up"):
                obj.clean_up()
            delattr(self, attr)


# -------------------------------------------------------------------------
#
# Helper functions
#
# -------------------------------------------------------------------------
def set_titles(window, title, text, msg=None):
    if title:
        title.set_text('<span weight="bold" size="larger">%s</span>' % text)
        title.set_use_markup(True)
    if msg:
        window.set_title("%s - Gramps" % msg)
    elif text:
        window.set_title("%s - Gramps" % text)
    window.set_icon_from_file(ICON)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2010       Michiel D. Nauta
# Copyright (C) 2011       Tim G L Lyons
# Copyright (C) 2013       Doug Blank <doug.blank@gmail.com>
# Copyright (C) 2017       Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Source object for Gramps.
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..const import GRAMPS_LOCALE as glocale
from .attrbase import SrcAttributeBase
from .citationbase import IndirectCitationBase
from .const import DIFFERENT, EQUAL, IDENTICAL
from .mediabase import MediaBase
from .notebase import NoteBase
from .primaryobj import PrimaryObject
from .reporef import RepoRef
from .tagbase import TagBase

_ = glocale.translation.gettext


# -------------------------------------------------------------------------
#
# Source
#
# -------------------------------------------------------------------------
class Source(
    MediaBase, NoteBase, SrcAttributeBase, IndirectCitationBase, PrimaryObject
):
    """
    A record of a source of information.
    """

    def __init__(self):
        """Create a new Source instance."""
        PrimaryObject.__init__(self)
        MediaBase.__init__(self)
        NoteBase.__init__(self)
        SrcAttributeBase.__init__(self)
        self.title = ""
        self.author = ""
        self.pubinfo = ""
        self.abbrev = ""
        self.reporef_list = []

    def serialize(self):
        """
        Convert the object to a serialized tuple of data.
        """
        return (
            self.handle,  # 0
            self.gramps_id,  # 1
            str(self.title),  # 2
            str(self.author),  # 3
            str(self.pubinfo),  # 4
            NoteBase.serialize(self),  # 5
            MediaBase.serialize(self),  # 6
            str(self.abbrev),  # 7
            self.change,  # 8
            SrcAttributeBase.serialize(self),  # 9
            [rr.serialize() for rr in self.reporef_list],  # 10
            TagBase.serialize(self),  # 11
            self.private,
        )  # 12

    @classmethod
    def get_schema(cls):
        """
        Returns the JSON Schema for this class.

        :returns: Returns a dict containing the schema.
        :rtype: dict
        """
        # pylint: disable=import-outside-toplevel
        from .mediaref import MediaRef
        from .srcattribute import SrcAttribute

        return {
            "type": "object",
            "title": _("Source"),
            "properties": {
                "_class": {"enum": [cls.__name__]},
                "handle": {
                    "type": "string",
                    "maxLength": 50,
                    "title": _("Handle"),
                },
                "gramps_id": {"type": "string", "title": _("Gramps ID")},
                "title": {"type": "string", "title": _("Title")},
                "author": {"type": "string", "title": _("Author")},
                "pubinfo": {"type": "string", "title": _("Publication info")},
                "note_list": {
                    "type": "array",
                    "items": {"type": "string", "maxLength": 50},
                    "title": _("Notes"),
                },
                "media_list": {
                    "type": "array",
                    "items": MediaRef.get_schema(),
                    "title": _("Media"),
                },
                "abbrev": {"type": "string", "title": _("Abbreviation")},
                "change": {"type": "integer", "title": _("Last changed")},
                "attribute_list": {
                    "type": "array",
                    "items": SrcAttribute.get_schema(),
                    "title": _("Source Attributes"),
                },
                "reporef_list": {
                    "type": "array",
                    "items": RepoRef.get_schema(),
                    "title": _("Repositories"),
                },
                "tag_list": {
                    "type": "array",
                    "items": {"type": "string", "maxLength": 50},
                    "title": _("Tags"),
                },
                "private": {"type": "boolean", "title": _("Private")},
            },
        }

    def unserialize(self, data):
        """
        Convert the data held in a tuple created by the serialize method
        back into the data in a Source structure.
        """
        (
            self.handle,  #  0
            self.gramps_id,  #  1
            self.title,  #  2
            self.author,  #  3
            self.pubinfo,  #  4
            note_list,  #  5
            media_list,  #  6
            self.abbrev,  #  7
            self.change,  #  8
            srcattr_list,  #  9
            reporef_list,  #  10
            tag_list,  #  11
            self.private,  #  12
        ) = data

        NoteBase.unserialize(self, note_list)
        MediaBase.unserialize(self, media_list)
        TagBase.unserialize(self, tag_list)
        SrcAttributeBase.unserialize(self, srcattr_list)
        self.reporef_list = [RepoRef().unserialize(item) for item in reporef_list]
        return self

    def _has_handle_reference(self, classname, handle):
        """
        Return True if the object has reference to a given handle of given
        primary object type.

        :param classname: The name of the primary object class.
        :type classname: str
        :param handle: The handle to be checked.
        :type handle: str
        :returns: Returns whether the object has reference to this handle of
                  this object type.
        :rtype: bool
        """
        if classname == "Repository":
            return handle in [ref.ref for ref in self.reporef_list]
        return False

    def _remove_handle_references(self, classname, handle_list):
        """
        Remove all references in this object to object handles in the list.

        :param classname: The name of the primary object class.
        :type classname: str
        :param handle_list: The list of handles to be removed.
        :type handle_list: str
        """
        if classname == "Repository":
            new_list = [ref for ref in self.reporef_list if ref.ref not in handle_list]
            self.reporef_list = new_list

    def _replace_handle_reference(self, classname, old_handle, new_handle):
        """
        Replace all references to old handle with those to the new handle.

        :param classname: The name of the primary object class.
        :type classname: str
        :param old_handle: The handle to be replaced.
        :type old_handle: str
        :param new_handle: The handle to replace the old one with.
        :type new_handle: str
        """
        if classname == "Repository":
            handle_list = [ref.ref for ref in self.reporef_list]
            while old_handle in handle_list:
                idx = handle_list.index(old_handle)
                self.reporef_list[idx].ref = new_handle
                handle_list[idx] = ""

    def get_text_data_list(self):
        """
        Return the list of all textual attributes of the object.

        :returns: Returns the list of all textual attributes of the object.
        :rtype: list
        """
        return [
            self.title,
            self.author,
            self.pubinfo,
            self.abbrev,
            self.gramps_id,
        ]

    def get_text_data_child_list(self):
        """
        Return the list of child objects that may carry textual data.

        :returns: Returns the list of child objects that may carry textual data.
        :rtype: list
        """
        return self.media_list + self.reporef_list + self.attribute_list

    def get_citation_child_list(self):
        """
        Return the list of child secondary objects that may refer citations.

        :returns: Returns the list of child secondary child objects that may
                  refer citations.
        :rtype: list
        """
        return self.media_list

    def get_note_child_list(self):
        """
        Return the list of child secondary objects that may refer notes.

        :returns: Returns the list of child secondary child objects that may
                  refer notes.
        :rtype: list
        """
        return self.media_list + self.reporef_list

    def get_handle_referents(self):
        """
        Return the list of child objects which may, directly or through
        their children, reference primary objects.

        :returns: Returns the list of objects referencing primary objects.
        :rtype: list
        """
        return self.get_citation_child_list() + self.reporef_list

    def get_referenced_handles(self):
        """
        Return the list of (classname, handle) tuples for all directly
        referenced primary objects.

        :returns: List of (classname, handle) tuples for referenced objects.
        :rtype: list
        """
        return self.get_referenced_note_handles() + self.get_referenced_tag_handles()

    def merge(self, acquisition):
        """
        Merge the content of acquisition into this source.

        :param acquisition: The source to merge with the present source.
        :type acquisition: Source
        """
        self._merge_privacy(acquisition)
        self._merge_note_list(acquisition)
        self._merge_media_list(acquisition)
        self._merge_tag_list(acquisition)
        self._merge_attribute_list(acquisition)
        self._merge_reporef_list(acquisition)

    def set_title(self, title):
        """
        Set the descriptive title of the Source object.

        :param title: descriptive title to assign to the Source
        :type title: str
        """
        self.title = title

    def get_title(self):
        """
        Return the descriptive title of the Place object.

        :returns: Returns the descriptive title of the Place
        :rtype: str
        """
        return self.title

    def set_author(self, author):
        """Set the author of the Source."""
        self.author = author

    def get_author(self):
        """Return the author of the Source."""
        return self.author

    def set_publication_info(self, text):
        """Set the publication information of the Source."""
        self.pubinfo = text

    def get_publication_info(self):
        """Return the publication information of the Source."""
        return self.pubinfo

    def set_abbreviation(self, abbrev):
        """Set the title abbreviation of the Source."""
        self.abbrev = abbrev

    def get_abbreviation(self):
        """Return the title abbreviation of the Source."""
        return self.abbrev

    def add_repo_reference(self, repo_ref):
        """
        Add a :class:`~.reporef.RepoRef` instance to the Source's reporef list.

        :param repo_ref: :class:`~.reporef.RepoRef` instance to be added to the
                         object's reporef list.
        :type repo_ref: :class:`~.reporef.RepoRef`
        """
        self.reporef_list.append(repo_ref)

    def get_reporef_list(self):
        """
        Return the list of :class:`~.reporef.RepoRef` instances associated with
        the Source.

        :returns: list of :class:`~.reporef.RepoRef` instances associated with
                  the Source
        :rtype: list
        """
        return self.reporef_list

    def set_reporef_list(self, reporef_list):
        """
        Set the list of :class:`~.reporef.RepoRef` instances associated with
        the Source. It replaces the previous list.

        :param reporef_list: list of :class:`~.reporef.RepoRef` instances to be
                             assigned to the Source.
        :type reporef_list: list
        """
        self.reporef_list = reporef_list

    def _merge_reporef_list(self, acquisition):
        """
        Merge the list of repository references from acquisition with our own.

        :param acquisition: the repository references list of this object will
                            be merged with the current repository references
                            list.
        :type acquisition: RepoRef
        """
        reporef_list = self.reporef_list[:]
        for addendum in acquisition.get_reporef_list():
            for reporef in reporef_list:
                equi = reporef.is_equivalent(addendum)
                if equi == IDENTICAL:
                    break
                if equi == EQUAL:
                    reporef.merge(addendum)
                    break
            else:
                self.reporef_list.append(addendum)

    def has_repo_reference(self, repo_handle):
        """
        Return True if the Source has reference to this Repository handle.

        :param repo_handle: The Repository handle to be checked.
        :type repo_handle: str
        :returns: Returns whether the Source has reference to this Repository
                  handle.
        :rtype: bool
        """
        return repo_handle in [repo_ref.ref for repo_ref in self.reporef_list]

    def remove_repo_references(self, repo_handle_list):
        """
        Remove references to all Repository handles in the list.

        :param repo_handle_list: The list of Repository handles to be removed.
        :type repo_handle_list: list
        """
        new_reporef_list = [
            repo_ref
            for repo_ref in self.reporef_list
            if repo_ref.ref not in repo_handle_list
        ]
        self.reporef_list = new_reporef_list

    def replace_repo_references(self, old_handle, new_handle):
        """
        Replace all references to old Repository handle with the new handle
        and merge equivalent entries.

        :param old_handle: The Repository handle to be replaced.
        :type old_handle: str
        :param new_handle: The Repository handle to replace the old one with.
        :type new_handle: str
        """
        refs_list = [repo_ref.ref for repo_ref in self.reporef_list]
        new_ref = None
        if new_handle in refs_list:
            new_ref = self.reporef_list[refs_list.index(new_handle)]
        n_replace = refs_list.count(old_handle)
        for _ in range(n_replace):
            idx = refs_list.index(old_handle)
            self.reporef_list[idx].ref = new_handle
            refs_list[idx] = new_handle
            if new_ref:
                repo_ref = self.reporef_list[idx]
                equi = new_ref.is_equivalent(repo_ref)
                if equi != DIFFERENT:
                    if equi == EQUAL:
                        new_ref.merge(repo_ref)
                    self.reporef_list.pop(idx)
                    refs_list.pop(idx)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2008-2010  Craig J. Anderson
# Copyright (C) 2014       Paul Franklin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""Reports/Graphical Reports/Tree_Base"""

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext
from gramps.gen.plug.report import utils
from gramps.plugins.lib.libsubstkeyword import SubstKeywords
from gramps.gen.plug.docgen import IndexMark, INDEX_TYPE_TOC

PT2CM = utils.pt2cm


# ------------------------------------------------------------------------
#
# Class Calc_Lines
#
# ------------------------------------------------------------------------
class CalcLines:
    """wrapper for libsubstkeyword and added functionality for
    replacements.

    Receive:  Individual and family handle, and display format [string]
    return: [Text] ready for a box.
    """

    def __init__(self, dbase, repl, locale, name_displayer):
        self.database = dbase
        self.display_repl = repl
        # self.default_string = default_str
        self._locale = locale
        self._nd = name_displayer

    def calc_lines(self, _indi_handle, _fams_handle, workinglines):
        """
        In this pass we will:
        1. make our text and do our replacements
        2. remove any extra (unwanted) lines with the compres option
        """

        ####################
        # 1.1  Get our line information here
        subst = SubstKeywords(
            self.database, self._locale, self._nd, _indi_handle, _fams_handle
        )
        lines = subst.replace_and_clean(workinglines)

        ####################
        # 1.2  do our replacements
        lns = []
        for line in lines:
            for pair in self.display_repl:
                if pair.count("/") == 1:
                    repl = pair.split("/", 1)
                    line = line.replace(repl[0], repl[1])
            lns.append(line)

        return lns


# ------------------------------------------------------------------------
#
# Class Canvas/Pages
#
# ------------------------------------------------------------------------
class Page:
    """This class is a printable page.
    Offsets from the canvas, Page numbers
    boxes and lines
    """

    def __init__(self, canvas):
        # parts from canvas
        # self.doc = doc
        self.canvas = canvas

        # parts about the page
        self.page_x_offset = 0
        self.page_y_offset = 0
        self.x_page_num = 0
        self.y_page_num = 0
        self.boxes = []  # All object must derive from BoxBase
        self.lines = []  # All must derive form Linebase
        self.note = None

    def is_blank(self):
        """Am I a blank page?  Notes and Titles are boxes too"""
        return self.boxes == [] and self.lines == []

    def add_box(self, box):
        """The box must derive from class Box_Base(Object):"""
        self.boxes.append(box)
        box.page = self

    def add_line(self, line):
        """Add a line onto this page"""
        self.lines.append(line)

    def draw_border(self, line_name):
        doc = self.canvas.doc
        if self.y_page_num == 0:
            doc.draw_line(line_name, 0, 0, doc.get_usable_width(), 0)
        if self.x_page_num == 0:
            doc.draw_line(line_name, 0, 0, 0, doc.get_usable_height())
        if self.y_page_num == self.canvas.y_pages - 1:
            doc.draw_line(
                line_name,
                0,
                doc.get_usable_height(),
                doc.get_usable_width(),
                doc.get_usable_height(),
            )
        if self.x_page_num == self.canvas.x_pages - 1:
            doc.draw_line(
                line_name,
                doc.get_usable_width(),
                0,
                doc.get_usable_width(),
                doc.get_usable_height(),
            )

    def display(self):
        """Display all boxes and lines that are on this page"""
        for box in self.boxes:
            box.display()
        for line in self.lines:
            line.display(self)


class Canvas(Page):
    """The Canvas is two things.
    The all in one canvas.  a canvas is a page of unlimited size
    a group of pages.  each page is set is size and shows only a
     part of what is on the entire canvas
    """

    def __init__(self, doc, report_opts):
        Page.__init__(self, self)
        self.doc = doc
        self.report_opts = report_opts

        # How many pages are there in the report.  one more than real.
        self.x_pages = 1
        self.y_pages = 1
        self.__pages = {(0, 0): self}  # set page 0,0 to me.
        self.__fonts = {}  # keep a list of fonts so we don't have to lookup.
        self.title = None
        self.note = None

    def __new_page(self, x_page, y_page, x_offset, y_offset):
        """Make a new page.  This will only happen if we are
        paginating (making new pages to hold parts of the canvas)"""
        if x_page >= self.x_pages:
            self.x_pages = x_page + 1
        new_page = Page(self)
        new_page.x_page_num = x_page
        new_page.y_page_num = y_page
        new_page.page_x_offset = x_offset
        new_page.page_y_offset = y_offset
        self.__pages[x_page, y_page] = new_page
        return new_page

    def sort_boxes_on_y_cm(self):
        """sorts the list of boxes on the canvas by .y_cm (top down)"""
        self.boxes.sort(key=lambda box: box.y_cm)

    def add_title(self, title):
        """The title must derive from class TitleBox(BoxBase):"""
        self.title = title
        self.title.cm_y = self.report_opts.littleoffset

    def add_note(self, note):
        """The note must derive from class NoteBox(BoxBase, NoteType)"""
        self.note = note
        self.set_box_height_width(self.note)

    def __get_font(self, box):
        """returns the font used by a box.  makes a list of all seen fonts
        to be faster.  If a new is found, run through the process to get it"""
        if box.boxstr not in self.__fonts:
            style_sheet = self.doc.get_style_sheet()
            style_name = style_sheet.get_draw_style(box.boxstr)
            style_name = style_name.get_paragraph_style()
            self.__fonts[box.boxstr] = style_sheet.get_paragraph_style(
                style_name
            ).get_font()

        return self.__fonts[box.boxstr]

    def get_report_height_width(self):
        """returns the (max width, max height) of the report
        This does not take into account any shadows"""
        max_width = 0
        max_height = 0
        for box in self.boxes:
            tmp = box.x_cm + box.width
            if tmp > max_width:
                max_width = tmp
            tmp = box.y_cm + box.height
            if tmp > max_height:
                max_height = tmp
        max_width += self.report_opts.box_shadow
        max_width += self.report_opts.littleoffset
        max_height += self.report_opts.box_shadow
        max_height += self.report_opts.littleoffset
        return (max_width, max_height)

    def __scale_canvas(self, scale_amount):
        """scales everything up/down depending upon scale_amount"""
        self.report_opts.scale_everything(scale_amount)
        self.title.scale(scale_amount)
        if self.note is not None:
            self.note.scale(scale_amount)
        # scale down everyone!
        for box in self.boxes:
            box.scale(scale_amount)

    def set_box_height_width(self, box):
        """Sets the .width and .height of a box."""
        if box.boxstr == "None":
            box.height = box.width = 0
            return

        font = self.__get_font(box)
        #####################
        # Get the width
        for line in box.text:
            width = self.doc.string_width(font, line)
            width = PT2CM(width)
            if width > box.width:
                box.width = width

        #####################
        # Get the height
        height = len(box.text) * font.get_size() * 1.5
        height += 1.0 / 2.0 * font.get_size()  # funny number(s) based upon font.
        box.height = PT2CM(height)

    def page_count(self, incblank):
        count = 0
        if incblank:
            return self.y_pages * self.x_pages

        for y_p in range(self.y_pages):
            for x_p in range(self.x_pages):
                if (x_p, y_p) in self.__pages:
                    count += 1
        return count

    def page_iter_gen(self, incblank):
        """generate the pages of the report.  do so in a left to right
        up down approach.  incblank asks to include blank pages"""
        blank = Page(self)
        for y_p in range(self.y_pages):
            for x_p in range(self.x_pages):
                if (x_p, y_p) in self.__pages:
                    yield self.__pages[(x_p, y_p)]
                else:
                    if incblank:
                        blank.x_page_num = x_p
                        blank.y_page_num = y_p
                        yield blank

    def __add_box_to_page(self, x_page, y_page, x_offset, y_offset, box):
        """adds a box to a page.  If the page is not there, make it first"""
        if (x_page, y_page) not in self.__pages:
            # Add the new page into the dictionary
            self.__new_page(x_page, y_page, x_offset, y_offset)

        # Add the box into the page
        self.__pages[x_page, self.y_pages - 1].add_box(box)

    def scale_report(self, one_page, scale_to_width, scale_to_height):
        """We have a report in its full size (on the canvas
        and pages to print on.  scale one or both as needed/desired.

        - one_page, boolean.  Whether to make the page(or parts of) the size
            of the report
        - scale_to_width, boolean.  Scale the report width to the page size?
        - scale_to_height, boolean.  Scale the report height to page size?
        """

        if scale_to_width or scale_to_height:
            max_width, max_height = self.canvas.get_report_height_width()
            # max_width  += self.report_opts.littleoffset
            # max_height += self.report_opts.littleoffset

        """
        calc - Calculate the scale amount (if any).
          <1 everything is smaller to fit on the page
          1 == no scaling
          >1 make everything bigger to fill out the page
        """
        scale = 1
        scaled_report_to = None

        #####################
        # scale the report option - width
        if scale_to_width:
            # Check the width of the title
            title_width = self.title.width
            title_width += self.report_opts.littleoffset * 2

            max_width = max(title_width, max_width)

            # This will be our base amount and
            # then we will decrease only as needed from here.

            scale = self.doc.get_usable_width() / max_width
            scaled_report_to = "width"

        #####################
        # scale the report option - height
        if scale_to_height:
            tmp = self.doc.get_usable_height() / max_height
            if not scale_to_width or tmp < scale:
                scale = tmp
                scaled_report_to = "height"

        # Now I have the scale amount
        if scale != 1:  # scale everything on the canvas
            self.__scale_canvas(scale)

        #####################
        # Scale the page option
        if one_page:
            # user wants PAGE to be the size of the report.
            size = self.doc.paper.get_size()
            size.name = "custom"

            max_width, max_height = self.canvas.get_report_height_width()

            if scaled_report_to != "width":
                # calculate the width of the report
                # max_width  += self.report_opts.littleoffset
                max_width += self.doc.paper.get_left_margin()
                max_width += self.doc.paper.get_right_margin()

                # calculate the width of the title
                title_width = self.canvas.title.width
                title_width += self.doc.paper.get_left_margin()
                title_width += self.doc.paper.get_right_margin()
                title_width += self.report_opts.littleoffset
                max_width = max(title_width, max_width)

                size.set_width(max_width)

            if scaled_report_to != "height":
                # calculate the height of the report
                max_height += self.doc.paper.get_top_margin()
                max_height += self.doc.paper.get_bottom_margin()
                # max_height += self.report_opts.littleoffset
                size.set_height(max_height)

        return scale

    def __paginate_x_offsets(self, colsperpage):
        """Go through the boxes and get the x page offsets"""
        # fix soon.  should not use .level
        liloffset = self.report_opts.littleoffset
        x_page_offsets = {0: 0}  # change me to [] ???
        for box in self.boxes:
            x_index = box.level[0]
            x_page = x_index // colsperpage
            if x_page not in x_page_offsets and x_index % colsperpage == 0:
                x_page_offsets[x_page] = box.x_cm - liloffset
                if x_page >= self.x_pages:
                    self.x_pages = x_page + 1
        return x_page_offsets

    def __paginate_y_pages(self, colsperpage, x_page_offsets):
        """Go through the boxes and put each one in a page
        note that the self.boxes needs to be sorted by .y_cm"""
        page_y_top = [0]
        page_y_height = [self.doc.get_usable_height()]
        liloffset = self.report_opts.littleoffset

        for box in self.boxes:
            # check to see if this box cross over to the next (y) page
            height = box.y_cm + liloffset + box.height  # + box.shadow/2

            if height > page_y_height[-1]:
                # we went off the end
                page_y_height.append(box.y_cm - liloffset + page_y_height[0])
                page_y_top.append(box.y_cm - liloffset)
                self.y_pages = len(page_y_height)

            # Calculate my (x) page
            # fix soon.  should not use .level
            x_page = box.level[0] // colsperpage

            self.__add_box_to_page(
                x_page,
                self.y_pages - 1,
                x_page_offsets[x_page],
                page_y_top[self.y_pages - 1],
                box,
            )
            # if not self.__pages.has_key((x_page, self.y_pages-1)):
            #    #Add the new page into the dictionary
            #    self.__new_page(x_page, self.y_pages-1,
            #                    )
            #
            ##Add the box into the page
            # self.__pages[x_page, self.y_pages-1].add_box(box)
        return page_y_top

    def __paginate_note(self, x_page_offsets, page_y_top):
        """Put the note on first.  it can be overwritten by other
        boxes but it CAN NOT overwrite a box."""
        x_page, y_page = self.note.set_on_page(self)
        if (x_page, y_page) not in self.__pages:
            # Add the new page into the dictionary
            self.__new_page(x_page, y_page, x_page_offsets[x_page], page_y_top[y_page])
        # Add the box into the page
        self.__pages[x_page, y_page].boxes.insert(0, self.note)
        self.note.doc = self.doc
        self.note.page = self

    def __paginate_lines(self, x_page_offsets, page_y_top):
        """Step three go through the lines and put each in page(s)"""
        for box1 in self.boxes:
            if not box1.line_to:
                continue

            line = box1.line_to

            pages = [box1.page.y_page_num]

            end = line.start + line.end

            x_page = box1.page.x_page_num
            start_y_page = end[0].page.y_page_num
            end_y_page = end[0].page.y_page_num

            for box in end:
                y_page = box.page.y_page_num
                if y_page not in pages:
                    if (x_page, y_page) not in self.__pages:
                        # Add the new page into the dictionary
                        self.__new_page(
                            x_page, y_page, x_page_offsets[x_page], page_y_top[y_page]
                        )
                    self.__pages[x_page, y_page].add_line(box1.line_to)
                    pages.append(y_page)

                if y_page < start_y_page:
                    start_y_page = y_page
                if y_page > end_y_page:
                    end_y_page = y_page

            # if len(end) = 2 & end[0].y_page = 0 & end[1].y_page = 4
            # the line will not print on y_pages 1,2,3.  Fix that here.
            # x_page = start_x_page
            for y_page in range(start_y_page, end_y_page + 1):
                if y_page not in pages:
                    if (x_page, y_page) not in self.__pages:
                        # Add the new page into the dictionary
                        self.__new_page(
                            x_page, y_page, x_page_offsets[x_page], page_y_top[y_page]
                        )
                    self.__pages[x_page, y_page].add_line(box1.line_to)

    def __paginate_title(self, x_page_offsets):
        # step four work with the title
        if self.title.boxstr == "None":
            return
        # x_page_offsets[page] tells me the widths I can use
        if len(x_page_offsets) > 1:
            if self.title.mark_text and not self.title.text:
                self.title.width = self.doc.get_usable_width()
                self.__pages[list(self.__pages.keys())[0]].add_box(self.title)
                return
            title_list = self.title.text.split(" ")
            title_font = self.__get_font(self.title)
            # space_width = PT2CM(self.doc.string_width(title_font," "))

            list_title = [title_list.pop(0)]
            while len(title_list):
                tmp = list_title[-1] + " " + title_list[0]
                if PT2CM(self.doc.string_width(title_font, tmp)) > x_page_offsets[1]:
                    list_title.append("")
                if list_title[-1] != "":
                    list_title[-1] += " "
                list_title[-1] += title_list.pop(0)

            start_page = int((len(x_page_offsets) - len(list_title)) / 2)
            for tmp in range(start_page):
                list_title.insert(0, "")
                list_title.append("")
            # one extra for security.  doesn't hurt.
            list_title.append("")

            x_page = 0
            for title in list_title:
                if title == "":
                    x_page += 1
                    continue
                if (x_page, 0) not in self.__pages:
                    # Add the new page into the dictionary
                    self.__new_page(x_page, 0, x_page_offsets[1], 0)

                title_part = TitleBox(self.doc, self.title.boxstr)
                title_part.text = list_title[x_page]
                title_part.width = x_page_offsets[1]

                # Add the box into the page
                self.__pages[x_page, 0].add_box(title_part)
                x_page = x_page + 1
        else:
            self.title.width = self.doc.get_usable_width()
            self.__pages[0, 0].add_box(self.title)

    def __paginate(self, colsperpage):
        """take the boxes on the canvas and put them into separate pages.
        The boxes need to be sorted by y_cm"""
        liloffset = self.report_opts.littleoffset
        self.__pages = {}
        x_page_offsets = self.__paginate_x_offsets(colsperpage)
        page_y_top = self.__paginate_y_pages(colsperpage, x_page_offsets)

        if self.note is not None:
            self.__paginate_note(x_page_offsets, page_y_top)
        self.__paginate_lines(x_page_offsets, page_y_top)
        self.__paginate_title(x_page_offsets)

    def paginate(self, colsperpage, one_page_report):
        """self.boxes must be sorted by box.y_cm for this to work."""
        if one_page_report:
            # self.canvas.add_box(self.canvas.title)
            title_part = TitleBox(self.doc, self.title.boxstr)
            title_part.text = self.title.text
            title_part.width = self.doc.get_usable_width()
            self.add_box(title_part)

            if self.note is not None:
                self.note.set_on_page(self)
                self.boxes.insert(0, self.note)
                self.note.doc = self.doc
                self.note.page = self
        else:
            self.__paginate(colsperpage)


# ------------------------------------------------------------------------
#
# Class Box_Base
#
# ------------------------------------------------------------------------
class BoxBase:
    """boxes are always in/on a Page
    Needed to print are: boxstr, text, x_cm, y_cm, width, height
    """

    def __init__(self):
        self.page = None

        #'None' will cause an error.  Sub-classes will init
        self.boxstr = "None"
        self.text = ""
        # level requires ...
        # (#  - which column am I in  (zero based)
        # ,#  - Am I a (0)direct descendant/ancestor or (>0)other
        # , ) - anything else the report needs to run
        self.__mark = None  # Database person object
        self.level = (0, 0)
        self.x_cm = 0.0
        self.y_cm = 0.0
        self.width = 0.0
        self.height = 0.0
        self.line_to = None
        # if text in TOC needs to be different from text, set mark_text
        self.mark_text = None

    def scale(self, scale_amount):
        """Scale the amounts"""
        self.x_cm *= scale_amount
        self.y_cm *= scale_amount
        self.width *= scale_amount
        self.height *= scale_amount

    def add_mark(self, database, person):
        self.__mark = utils.get_person_mark(database, person)

    def display(self):
        """display the box accounting for page x, y offsets
        Ignore any box with 'None' is boxstr"""
        if self.boxstr == "None":
            return

        doc = self.page.canvas.doc
        report_opts = self.page.canvas.report_opts
        text = "\n".join(self.text)
        xbegin = self.x_cm - self.page.page_x_offset
        ybegin = self.y_cm - self.page.page_y_offset

        doc.draw_box(
            self.boxstr, text, xbegin, ybegin, self.width, self.height, self.__mark
        )

        # I am responsible for my own lines. Do them here.
        if self.line_to:
            # draw my line out here.
            self.line_to.display(self.page)
        if (
            self.page.x_page_num > 0
            and self.level[1] == 0
            and xbegin < report_opts.littleoffset * 2
        ):
            # I am a child on the first column
            yme = ybegin + self.height / 2
            doc.draw_line(report_opts.line_str, 0, yme, xbegin, yme)


class TitleNoDisplay(BoxBase):
    """
    Holds information about the Title that will print on a TOC
    and NOT on the report
    """

    def __init__(self, doc, boxstr):
        """initialize the title box"""
        BoxBase.__init__(self)
        self.doc = doc
        self.boxstr = boxstr

    def set_box_height_width(self):
        self.width = self.height = 0

    def display(self):
        """display the title box."""
        # Set up the Table of Contents here
        if self.mark_text is None:
            mark = IndexMark(self.text, INDEX_TYPE_TOC, 1)
        else:
            mark = IndexMark(self.mark_text, INDEX_TYPE_TOC, 1)
        self.doc.center_text(self.boxstr, "", 0, -100, mark)


class TitleBox(BoxBase):
    """
    Holds information about the Title that will print on a page
    """

    def __init__(self, doc, boxstr):
        """initialize the title box"""
        BoxBase.__init__(self)
        self.doc = doc
        self.boxstr = boxstr
        if boxstr == "None":
            return

        style_sheet = self.doc.get_style_sheet()
        style_name = style_sheet.get_draw_style(self.boxstr)
        style_name = style_name.get_paragraph_style()
        self.font = style_sheet.get_paragraph_style(style_name).get_font()

    def set_box_height_width(self):
        if self.boxstr == "None":
            return
        # fix me. width should be the printable area
        self.width = PT2CM(self.doc.string_width(self.font, self.text))
        self.height = PT2CM(self.font.get_size() * 2)

    def _get_names(self, persons, name_displayer):
        """A helper function that receives a list of persons and
        returns their names in a list"""
        return [name_displayer.display(person) for person in persons]

    def display(self):
        """display the title box."""
        if self.page.y_page_num or self.boxstr == "None":
            return

        # Set up the Table of Contents here
        mark = IndexMark(self.text, INDEX_TYPE_TOC, 1)

        if self.text:
            self.doc.center_text(
                self.boxstr, self.text, self.width / 2, self.y_cm, mark
            )


class PageNumberBox(BoxBase):
    """
    Calculates information about the page numbers that will print on a page
    do not put in a value for PageNumberBox.text.  this will be calculated for
    each page"""

    def __init__(self, doc, boxstr, locale):
        """initialize the page number box"""
        BoxBase.__init__(self)
        self.doc = doc
        self.boxstr = boxstr
        self._ = locale.translation.sgettext

    def __calc_position(self, page):
        """calculate where I am to print on the page(s)"""
        # Translators: needed for Arabic, ignore otherwise
        self.text = "(%d" + self._(",") + "%d)"

        style_sheet = self.doc.get_style_sheet()
        style_name = style_sheet.get_draw_style(self.boxstr)
        style_name = style_name.get_paragraph_style()
        font = style_sheet.get_paragraph_style(style_name).get_font()

        # calculate how much space is needed
        if page.canvas.x_pages > 10:
            tmp = "00"
        else:
            tmp = "0"
        if page.canvas.y_pages > 10:
            tmp += "00"
        else:
            tmp += "0"

        width = self.doc.string_width(font, "(,)" + tmp)
        width = PT2CM(width)
        self.width = width

        height = font.get_size() * 1.4
        height += 0.5 / 2.0 * font.get_size()  # funny number(s) based upon font.
        self.height = PT2CM(height)

        self.x_cm = self.doc.get_usable_width() - self.width
        self.y_cm = self.doc.get_usable_height() - self.height

    def display(self, page):
        """If this is the first time I am ran, get my position
        then display the page number"""
        if self.text == "":
            self.__calc_position(page)

        self.doc.draw_text(
            self.boxstr,
            self.text % (page.x_page_num + 1, page.y_page_num + 1),
            self.x_cm,
            self.y_cm,
        )


class NoteType:
    """Provide the different options (corners) to place the note"""

    TOPLEFT = 0
    TOPRIGHT = 1
    BOTTOMLEFT = 2
    BOTTOMRIGHT = 3

    _DEFAULT = BOTTOMRIGHT

    _DATAMAP = [
        (TOPLEFT, _("Top Left"), "Top Left"),
        (TOPRIGHT, _("Top Right"), "Top Right"),
        (BOTTOMLEFT, _("Bottom Left"), "Bottom Left"),
        (BOTTOMRIGHT, _("Bottom Right"), "Bottom Right"),
    ]

    def __init__(self, value, exclude=None):
        """initialize GrampsType"""
        self.value = value
        self.exclude = exclude
        # GrampsType.__init__(self, value)

    def note_locals(self, start=0):
        """generates an int of all the options"""
        for tuple in self._DATAMAP:
            if tuple[0] != self.exclude:
                yield tuple[0], tuple[1]


class NoteBox(BoxBase, NoteType):
    """Box that will hold the note to display on the page"""

    def __init__(self, doc, boxstr, box_corner, exclude=None):
        """initialize the NoteBox"""
        BoxBase.__init__(self)
        NoteType.__init__(self, box_corner, exclude)
        self.doc = doc
        self.boxstr = boxstr

    def set_on_page(self, canvas):
        """set the x_cm and y_cm given
        self.doc, leloffset, and title_height"""

        liloffset = canvas.report_opts.littleoffset
        # left or right side
        if self.value == NoteType.BOTTOMLEFT or self.value == NoteType.TOPLEFT:
            self.x_cm = liloffset
        else:
            self.x_cm = self.doc.get_usable_width() - self.width - liloffset
        # top or bottom
        if self.value == NoteType.TOPRIGHT or self.value == NoteType.TOPLEFT:
            self.y_cm = canvas.title.height + liloffset * 2
        else:
            self.y_cm = self.doc.get_usable_height() - self.height - liloffset

        """ helper function for canvas.paginate().
        return the (x, y) page I want to print on """
        if self.value == NoteType.TOPLEFT:
            return (0, 0)
        elif self.value == NoteType.TOPRIGHT:
            return (canvas.x_pages - 1, 0)
        elif self.value == NoteType.BOTTOMLEFT:
            return (0, canvas.y_pages - 1)
        elif self.value == NoteType.BOTTOMRIGHT:
            return (canvas.x_pages - 1, canvas.y_pages - 1)

    def display(self):
        """position the box and display"""
        title = self.page.canvas.title
        title_height = 0
        if title is not None:
            title_height = title.height
        text = "\n".join(self.text)
        self.doc.draw_box(
            self.boxstr, text, self.x_cm, self.y_cm, self.width, self.height
        )


# ------------------------------------------------------------------------
#
# Class Line_base
#
# ------------------------------------------------------------------------
class LineBase:
    """A simple line class.
    self.start is the box that we are drawing a line from
    self.end are the boxes that we are drawing lines to.
    """

    def __init__(self, start):
        # self.linestr = "None"
        self.start = [start]
        self.end = []

    def add_from(self, person):
        self.start.append(person)

    def add_to(self, person):
        """add destination boxes to draw this line to"""
        self.end.append(person)

    def display(self, page):
        """display the line.  left to right line.  one start, multiple end.
        page will tell us what parts of the line we can print"""
        if self.end == [] and len(self.start) == 1:
            return

        # y_cm and x_cm start points - take into account page offsets
        # yme = self.start.y_cm + self.start.height/2 - page.page_y_offset
        # if type(self.start) != type([]):
        #    self.start = [self.start]
        start = self.start[0]
        doc = start.page.canvas.doc
        report_opts = start.page.canvas.report_opts
        linestr = report_opts.line_str

        xbegin = start.x_cm + start.width - page.page_x_offset
        # out 3/4 of the way and x_cm end point(s)
        x34 = xbegin + (report_opts.col_width * 3 / 4)
        xend = xbegin + report_opts.col_width

        if x34 > 0:  # > 0 tell us we are printing on this page.
            usable_height = doc.get_usable_height()
            # 1 - Line from start box out
            for box in self.start:
                yme = box.y_cm + box.height / 2 - page.page_y_offset
                if box.page.y_page_num == page.y_page_num:
                    # and 0 < yme < usable_height and \
                    doc.draw_line(linestr, xbegin, yme, x34, yme)

            # 2 - vertical line
            mid = []
            for box in self.start + self.end:
                tmp = box.y_cm + box.height / 2
                mid.append(tmp)
            mid.sort()
            mid = [mid[0] - page.page_y_offset, mid[-1] - page.page_y_offset]
            if mid[0] < 0:
                mid[0] = 0
            if mid[1] > usable_height:
                mid[1] = usable_height
            # draw the connecting vertical line.
            doc.draw_line(linestr, x34, mid[0], x34, mid[1])
        else:
            x34 = 0

        # 3 - horizontal line(s)
        for box in self.end:
            if box.page.y_page_num == page.y_page_num:
                yme = box.y_cm + box.height / 2 - box.page.page_y_offset
                doc.draw_line(linestr, x34, yme, xend, yme)


# ------------------------------------------------------------------------
#
# Class report_options
#
# ------------------------------------------------------------------------
class ReportOptions:
    """
    A simple class to hold various report information
    Calculates
      the gap between persons,
      the column width, for lines,
      the left hand spacing for spouses (Descendant report only)
    """

    def __init__(self, doc, normal_font, normal_line):
        """initialize various report variables that are used"""
        self.box_pgap = PT2CM(1.25 * normal_font.get_size())  # gap between persons
        self.box_mgap = self.box_pgap / 2  # gap between marriage information
        self.box_shadow = PT2CM(normal_font.get_size()) * 0.6  # normal text
        self.spouse_offset = PT2CM(doc.string_width(normal_font, "0"))

        self.col_width = PT2CM(doc.string_width(normal_font, "(000,0)"))
        self.littleoffset = PT2CM(1)
        self.x_cm_cols = [self.littleoffset]

        self.line_str = normal_line

        # Things that will get added later
        self.max_box_width = 0
        self.max_box_height = 0

        self.scale = 1

    def scale_everything(self, amount):
        """Scale the amounts that are needed to generate a report"""
        self.scale = amount

        self.col_width *= amount
        self.littleoffset *= amount

        self.max_box_width *= amount  # box_width
        self.spouse_offset *= amount
        self.box_shadow *= amount


# =====================================
# "And Jesus said unto them ... , "If ye have faith as a grain of mustard
# seed, ye shall say unto this mountain, Remove hence to yonder place; and
# it shall remove; and nothing shall be impossible to you."
# Romans 1:17
